# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


from typing import AsyncGenerator, List, Tuple
from uuid import uuid4
from datetime import datetime
from secrets import token_urlsafe
from email.mime.text import MIMEText

from sqlalchemy.sql import select, and_, or_, func

from ..resources import Sessions, Config, DemoQueue

from ..decorators import (
    validate_community_type,
    validate_webhooks,
    validate_email
)


from ..templates import render_html

from ..tables import (
    community_table,
    scoreboard_total_table,
    scoreboard_table,
    user_table,
    api_key_table,
    statistic_table,
    server_table
)

from ..exceptions import (
    InvalidCommunity,
    InvalidSteamID,
    UserExists,
    ServerExists
)

from .models import (
    CommunityModel,
    MatchModel,
    ProfileModel,
    CommunityStatsModel, ProfileOverviewModel,
    PublicCommunityModel,
    ServerModel
)

from ..user import create_user

from .key import Key
from .match import Match
from .server import Server


class Community:
    def __init__(self, community_name: str) -> str:
        """Handles community interactions.

        Paramters
        ---------
        community_name: str
            ID of community.
        """

        self.community_name = community_name

    async def create_server(self, ip: str, port: int,
                            name: str) -> Tuple[ServerModel, Server]:
        """Used to create server.

        Parameters
        ----------
        ip : str
        port : int
        name : str

        Returns
        -------
        ServerModel
        Server

        Raises
        ------
        ServerExists
        """

        values = dict(
            community_name=self.community_name,
            ip=ip,
            port=port,
            name=name,
            players=0,
            max_players=0,
            map=None
        )

        try:
            await Sessions.database.execute(
                server_table.insert().values(
                    **values
                )
            )
        except Exception:
            raise ServerExists()
        else:
            return ServerModel(**values), self.server(ip, port)

    async def servers(self) -> AsyncGenerator[ServerModel, Server]:
        """Used to list servers.

        Yields
        -------
        ServerModel
        Server
        """

        query = server_table.select().where(
            server_table.c.community_name == self.community_name
        )

        async for row in Sessions.database.iterate(query):
            yield ServerModel(**row), self.server(row["ip"], row["port"])

    def server(self, ip: str, port: int) -> Server:
        """Used to interact with a server.

        Parameters
        ----------
        ip : str
        port : int

        Returns
        -------
        Server
        """

        return Server(ip, port, self.community_name)

    async def email(self, title: str, content: str,
                    link_href: str, link_text: str) -> None:
        """Used to send email to community owner.

        Parameters
        ----------
        title : str
        content : str
        link_href : str
        link_text : str
        """

        community = await self.get()

        message = MIMEText(render_html(
            "email.html",
            {
                "title": title,
                "content": content,
                "link": {
                    "href": link_href,
                    "text": link_text
                }
            }
        ), "html", "utf-8")

        message["From"] = Config.system_email
        message["To"] = community.email
        message["Subject"] = "{} - {}".format(
            title,
            (datetime.now()).strftime(Config.timestamp_format)
        )

        await Sessions.smtp.send_message(message)

    async def update_subscription_expire(self) -> None:
        await Sessions.database.execute(
            community_table.update().values(
                subscription_expires=datetime.now() +
                Config.subscription_length
            ).where(
                community_table.c.community_name == self.community_name
            )
        )

    async def billing_session(self) -> str:
        """Used to get a billing session.

        Returns
        -------
        str
        """

        league = await self.get()

        billing = await (
            Sessions.stripe.customer(league.customer_id)
        ).create_billing_session(
            "{}c/{}/owner".format(
                Config.frontend_url,
                league.community_name
            )
        )

        return billing.url

    async def checkout_session(self) -> str:
        """Used to get a checkout session.

        Returns
        -------
        str
            Session ID.
        """

        league = await self.get()

        owner_url = "{}c/{}/owner/".format(
            Config.frontend_url,
            league.community_name
        )

        return await (
            Sessions.stripe.customer(league.customer_id)
        ).create_checkout_session(
            cancel_url=owner_url + "cancel",
            success_url=owner_url + "success",
            price_id=Config.price_id
        )

    @validate_webhooks
    @validate_community_type
    @validate_email
    async def update(self, demos: bool = None,
                     community_type: str = None,
                     match_start_webhook: str = None,
                     round_end_webhook: str = None,
                     match_end_webhook: str = None,
                     allow_api_access: bool = None,
                     email: str = None) -> CommunityModel:
        """Used to update a community.

        Parameters
        ----------
        demos : bool, optional
            by default None
        community_type : str, optional
            by default None
        match_start_webhook : str, optional
            by default None
        round_end_webhook : str, optional
            by default None
        match_end_webhook : str, optional
            by default None
        allow_api_access : bool, optional
            by default None
        email : str, optional
            by default None

        Returns
        -------
        CommunityModel
        """

        values = {}

        if community_type is not None:
            values["community_type_id"] = Config.community_types[
                community_type
            ]

        if demos is not None:
            values["demos"] = demos

        if allow_api_access is not None:
            values["allow_api_access"] = allow_api_access

        if match_start_webhook:
            values["match_start_webhook"] = match_start_webhook

        if round_end_webhook:
            values["round_end_webhook"] = round_end_webhook

        if match_end_webhook:
            values["match_end_webhook"] = match_end_webhook

        if email:
            values["email"] = email

        if values:
            query = community_table.update().values(
                **values
            ).where(
                community_table.c.community_name == self.community_name
            )

            await Sessions.database.execute(query)

        return await self.get()

    async def stats(self) -> CommunityStatsModel:
        """Gets basic stats about community.

        Returns
        -------
        CommunityStatsModel
        """

        sub_active_matches = select([
            func.count().label("active_matches")
        ]).select_from(
            scoreboard_total_table
        ).where(
            and_(
                scoreboard_total_table.c.community_name == self.community_name,
                scoreboard_total_table.c.status == 1
            )
        ).alias("sub_active_matches")

        sub_stored_demos = select([
            func.count().label("stored_demos")
        ]).select_from(
            scoreboard_total_table
        ).where(
            and_(
                scoreboard_total_table.c.community_name == self.community_name,
                scoreboard_total_table.c.demo_status == 2
            )
        ).alias("sub_stored_demos")

        sub_total_matches = select([
            func.count().label("total_matches"),
        ]).select_from(
            scoreboard_total_table
        ).where(
            and_(
                scoreboard_total_table.c.community_name == self.community_name,
                scoreboard_total_table.c.status == 0
            )
        ).alias("sub_total_matches")

        sub_total_users = select([
            func.count().label("total_users"),
        ]).select_from(
            statistic_table
        ).where(
            statistic_table.c.community_name == self.community_name
        ).alias("sub_total_users")

        query = select([
            sub_total_users.c.total_users,
            sub_active_matches.c.active_matches,
            sub_stored_demos.c.stored_demos,
            sub_total_matches.c.total_matches
        ])

        return CommunityStatsModel(
            **await Sessions.database.fetch_one(query)
        )

    async def profile(self, steam_id: str) -> ProfileModel:
        """Get user profile.

        Parameters
        ----------
        steam_id : str

        Returns
        -------
        ProfileModel

        Raises
        ------
        InvalidSteamID
        """

        query = select([
            user_table.c.name,
            user_table.c.timestamp,
            statistic_table.c.steam_id,
            statistic_table.c.kills,
            statistic_table.c.headshots,
            statistic_table.c.assists,
            statistic_table.c.deaths,
            statistic_table.c.shots_fired,
            statistic_table.c.shots_hit,
            statistic_table.c.mvps
        ]).select_from(
            statistic_table.join(
                user_table,
                user_table.c.steam_id == statistic_table.c.steam_id
            )
        ).where(
            and_(
                statistic_table.c.steam_id == steam_id,
                statistic_table.c.community_name == self.community_name
            )
        )

        row = await Sessions.database.fetch_one(query=query)
        if row:
            return ProfileModel(**row)
        else:
            raise InvalidSteamID()

    async def players(self, search: str = None, page: int = 1,
                      limit: int = 8, desc: bool = True
                      ) -> AsyncGenerator[ProfileOverviewModel, None]:
        """Used to list community players.

        Parameters
        ----------
        search : str, optional
            by default None
        page : int, optional
            by default 1
        limit : int, optional
            by default 8
        desc : bool, optional
            by default True

        Yields
        ------
        ProfileOverviewModel
        """

        query = select([
            user_table.c.name,
            statistic_table.c.steam_id,
            statistic_table.c.kills,
            statistic_table.c.headshots,
            statistic_table.c.assists,
            statistic_table.c.deaths
        ]).select_from(
            statistic_table.join(
                user_table,
                user_table.c.steam_id == statistic_table.c.steam_id
            )
        ).where(
            statistic_table.c.community_name == self.community_name
        )

        if search:
            query = query.where(
                or_(
                    statistic_table.c.steam_id == search,
                    user_table.c.name.like("%{}%".format(search))
                )
            )

        query = query.order_by(
            statistic_table.c.kills.desc() if desc
            else statistic_table.c.kills.asc()
        ).limit(limit).offset((page - 1) * limit if page > 1 else 0)

        async for row in Sessions.database.iterate(query):
            yield ProfileOverviewModel(**row)

    async def delete_matches(self, matches: List[str]) -> None:
        """Used to bulk delete matches.

        Parameters
        ----------
        matches : List[str]
            List of match IDs to delete.
        """

        await Sessions.database.execute(
            scoreboard_total_table.delete().where(
                scoreboard_total_table.c.match_id ==
                scoreboard_table.c.match_id
            ).where(
                and_(
                    scoreboard_total_table.c.community_name ==
                    self.community_name,
                    scoreboard_total_table.c.match_id.in_(matches)
                )
            )
        )

        # Todo
        # Work out sqlalchemy left joining delete,
        # so i don't need this ugly mess.
        await Sessions.database.execute(
            scoreboard_total_table.delete().where(
                and_(
                    scoreboard_total_table.c.community_name
                    == self.community_name,
                    scoreboard_total_table.c.match_id.in_(matches)
                )
            )
        )

        if Config.upload_type:
            if self.community_name not in DemoQueue.matches:
                DemoQueue.matches[self.community_name] = matches
            else:
                DemoQueue.matches[self.community_name] += matches

    async def create_match(self, team_1_name: str, team_2_name: str,
                           team_1_side: int, team_2_side: int,
                           team_1_score: int, team_2_score: int,
                           map_name: str) -> Tuple[MatchModel, Match]:
        """Creates a match.

        Returns
        -------
        Match
            Used for interacting with matches.
        """

        match_id = str(uuid4())
        now = datetime.now()
        status = 1
        demo_status = 0

        query = scoreboard_total_table.insert().values(
            match_id=match_id,
            team_1_name=team_1_name,
            team_2_name=team_2_name,
            team_1_side=team_1_side,
            team_2_side=team_2_side,
            map=map_name,
            community_name=self.community_name,
            team_1_score=team_1_score,
            team_2_score=team_2_score,
            status=status,
            demo_status=demo_status,
            timestamp=now
        )

        await Sessions.database.execute(query=query)

        return MatchModel(
            match_id=match_id, timestamp=now, status=status,
            demo_status=demo_status, map=map_name, team_1_name=team_1_name,
            team_2_name=team_2_name, team_1_score=team_1_score,
            team_2_score=team_2_score, team_1_side=team_1_side,
            team_2_side=team_2_side, community_name=self.community_name
        ), self.match(match_id)

    def match(self, match_id) -> Match:
        """Handles interactions with a match

        Paramters
        ---------
        match_id: str
            ID of match
        """

        return Match(match_id, self.community_name)

    def key(self, api_key: str) -> Key:
        """Used to interact with key.

        Parameters
        ----------
        api_key : str

        Returns
        -------
        Key
        """

        return Key(api_key, self.community_name)

    async def user_to_key(self, steam_id: str) -> Tuple[str, Key]:
        """Used to get a API key from a steam ID.

        Parameters
        ----------
        steam_id : str

        Returns
        -------
        str
            The API key.
        Key
        """

        key = await Sessions.database.fetch_val(
            select([api_key_table.c.api_key]).select_from(
                api_key_table
            ).where(
                and_(
                    api_key_table.c.community_name == self.community_name,
                    api_key_table.c.owner_id == steam_id,
                    api_key_table.c.master == False  # noqa: E712
                )
            )
        )

        return key, self.key(key)

    async def create_key(self, steam_id: str) -> Tuple[str, Key]:
        """Used to create a API key for a user.

        Parameters
        ----------
        steam_id : str

        Returns
        -------
        str
            The API key.
        Key
        """

        key = token_urlsafe(24)

        try:
            await create_user(steam_id, "Unknown")
        except UserExists:
            pass

        await Sessions.database.execute(
            api_key_table.insert().values(
                api_key=key,
                owner_id=steam_id,
                timestamp=datetime.now(),
                community_name=self.community_name,
                master=False
            )
        )

        return key, self.key(key)

    async def regenerate_master(self) -> str:
        """Regenerates the master API key.
        """

        key = token_urlsafe(24)

        query = api_key_table.update().values(
            api_key=key,
            timestamp=datetime.now()
        ).where(
            and_(
                api_key_table.c.community_name == self.community_name,
                api_key_table.c.master == 1
            )
        )

        await Sessions.database.execute(query=query)

        return key

    async def exists(self) -> bool:
        """Checks if community exists with name.
        """

        query = select([func.count()]).select_from(community_table).where(
            community_table.c.community_name == self.community_name
        )

        return await Sessions.database.fetch_val(query=query) > 0

    async def matches(self, search: str = None,
                      page: int = 1, limit: int = 10, desc: bool = True,
                      require_scoreboard: bool = True
                      ) -> AsyncGenerator[MatchModel, Match]:
        """Lists matches.

        Paramters
        ---------
        search: str
        page: int
        limit: int
        desc: bool, optional
            by default True
        require_scoreboard : bool, optional
            If enabled scoreboard will need to be ready
            to pull match, by default True

        Yields
        ------
        MatchModel
            Holds basic match details.
        Match
            Used for interacting with a match.
        """

        query = select([
            scoreboard_total_table.c.match_id,
            scoreboard_total_table.c.timestamp,
            scoreboard_total_table.c.status,
            scoreboard_total_table.c.demo_status,
            scoreboard_total_table.c.map,
            scoreboard_total_table.c.team_1_name,
            scoreboard_total_table.c.team_2_name,
            scoreboard_total_table.c.team_1_score,
            scoreboard_total_table.c.team_2_score,
            scoreboard_total_table.c.team_1_side,
            scoreboard_total_table.c.team_2_side,
            scoreboard_total_table.c.community_name
        ])

        if search:
            like_search = "%{}%".format(search)

            query = query.select_from(
                scoreboard_total_table.join(
                    scoreboard_table,
                    scoreboard_table.c.match_id ==
                    scoreboard_total_table.c.match_id
                ).join(
                    user_table,
                    user_table.c.steam_id == scoreboard_table.c.steam_id
                )
            ).where(
                and_(
                    scoreboard_total_table.c.community_name ==
                    self.community_name,
                    or_(
                        scoreboard_total_table.c.match_id == search,
                        scoreboard_total_table.c.map.like(like_search),
                        scoreboard_total_table.c.team_1_name.like(
                            like_search),
                        scoreboard_total_table.c.team_2_name.like(
                            like_search),
                        user_table.c.name.like(like_search),
                        user_table.c.steam_id == search
                    )
                )
            )
        elif require_scoreboard:
            query = query.select_from(
                scoreboard_total_table.join(
                    scoreboard_table,
                    scoreboard_table.c.match_id ==
                    scoreboard_total_table.c.match_id
                )
            ).where(
                scoreboard_total_table.c.community_name
                == self.community_name,
            )
        else:
            query = query.select_from(
                scoreboard_total_table
            ).where(
                scoreboard_total_table.c.community_name
                == self.community_name,
            )

        query = query.distinct().order_by(
            scoreboard_total_table.c.timestamp.desc() if desc
            else scoreboard_total_table.c.timestamp.asc()
        ).limit(limit).offset((page - 1) * limit if page > 1 else 0)

        async for row in Sessions.database.iterate(query=query):
            yield MatchModel(**row), self.match(row["match_id"])

    async def public(self) -> PublicCommunityModel:
        """Used to get public data on a community.

        Returns
        -------
        PublicCommunityModel

        Raises
        ------
        InvalidCommunity
            Raised when community ID doesn't exist.
        """

        query = select([
            community_table.c.owner_id,
            community_table.c.disabled,
            community_table.c.community_name,
            community_table.c.timestamp,
            community_table.c.banned,
            community_table.c.allow_api_access
        ]).select_from(community_table).where(
            community_table.c.community_name == self.community_name
        )

        row = await Sessions.database.fetch_one(query)
        if row:
            return PublicCommunityModel(**row)
        else:
            raise InvalidCommunity()

    async def get(self) -> CommunityModel:
        """Gets base community details.

        Returns
        -------
        CommunityModel
            Holds community data.

        Raises
        ------
        InvalidCommunity
            Raised when community ID doesn't exist.
        """

        query = select([
            api_key_table.c.api_key,
            api_key_table.c.owner_id,
            community_table.c.disabled,
            community_table.c.banned,
            community_table.c.community_name,
            community_table.c.timestamp,
            community_table.c.allow_api_access,
            community_table.c.match_start_webhook,
            community_table.c.round_end_webhook,
            community_table.c.match_end_webhook,
            community_table.c.customer_id,
            community_table.c.email,
            community_table.c.subscription_expires
        ]).select_from(
            community_table.join(
                api_key_table,
                community_table.c.community_name ==
                api_key_table.c.community_name
            )
        ).where(
            and_(
                community_table.c.community_name == self.community_name,
                api_key_table.c.master == True  # noqa: E712
            )
        )

        row = await Sessions.database.fetch_one(query)
        if row:
            return CommunityModel(**row)
        else:
            raise InvalidCommunity()

    async def disable(self) -> None:
        """Disables a community.
        """

        query = community_table.update().where(
            community_table.c.community_name == self.community_name
        ).values(disabled=True)

        await Sessions.database.execute(query)



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


from operator import or_
from typing import Tuple
from sqlalchemy.sql import select, and_, func
from secrets import token_urlsafe
from datetime import datetime

from ..resources import Sessions, Config

from ..tables import (
    community_table,
    api_key_table,
)

from ..decorators import (
    validate_community_type,
    validate_webhooks,
    validate_community_name,
    validate_email
)

from ..user import create_user

from ..exceptions import (
    CommunityTaken,
    AlreadyCommunity,
    NoOwnership,
    InvalidAPIKey,
    UserExists,
    InvalidCustomer
)
from .models import (
    CommunityModel
)

from .community import Community


async def api_key_to_community(api_key: str) -> Tuple[Community, bool]:
    """Converts API key to community name.

    Raises
    ------
    InvalidAPIKey

    Returns
    -------
    str
        Community name
    bool
        If key is a master key.
    """

    query = select([
        community_table.c.community_name, api_key_table.c.master
    ]).select_from(
        community_table.join(
            api_key_table,
            api_key_table.c.community_name == community_table.c.community_name
        )
    ).where(
        and_(
            api_key_table.c.api_key == api_key,
            community_table.c.disabled == False,  # noqa: E712
            community_table.c.banned == False  # noqa: E712
        )
    ).where(
        or_(
            api_key_table.c.master == True,  # noqa: E712
            community_table.c.allow_api_access == True  # noqa: E712
        )
    )

    row = await Sessions.database.fetch_one(query=query)

    if row:
        return Community(row["community_name"]), bool(row["master"])
    else:
        raise InvalidAPIKey()


async def stripe_customer_to_community(customer_id: str) -> Community:
    """Converts stripe customer ID to community.

    Parameters
    ----------
    customer_id : str

    Returns
    -------
    Community

    Raises
    ------
    InvalidCustomer
    """

    query = select([community_table.c.community_name]).select_from(
        community_table
    ).where(
        community_table.c.customer_id == customer_id
    )

    community_name = await Sessions.database.fetch_val(query)
    if not community_name:
        raise InvalidCustomer()

    return Community(community_name)


async def get_community_from_owner(steam_id: str
                                   ) -> Tuple[Community, bool, bool]:
    """Used to get a community from owner.

    Parameters
    ----------
    steam_id : str

    Returns
    -------
    Community
    bool
        If banned.
    bool
        If active subscription.

    Raises
    ------
    NoOwnership
        Raised when steam id doesn't own any communties.
    """

    query = select([
        community_table.c.community_name,
        community_table.c.banned,
        community_table.c.subscription_expires
    ]).select_from(
        community_table
    ).where(
        and_(
            community_table.c.owner_id == steam_id,
            community_table.c.disabled == False  # noqa: E712
        )
    )

    row = await Sessions.database.fetch_one(query)
    if row:
        return (
            Community(row["community_name"]),
            row["banned"],
            row["subscription_expires"] >= datetime.now()
            if row["subscription_expires"] else False
        )
    else:
        raise NoOwnership()


async def owner_exists(steam_id: str) -> bool:
    """Checks if given steam_id owns a community
    """

    query = select([func.count()]).select_from(community_table).where(
        and_(
            community_table.c.owner_id == steam_id,
            community_table.c.disabled == False  # noqa: E712
        )
    )

    return await Sessions.database.fetch_val(query=query) > 0


@validate_webhooks
@validate_community_name
@validate_community_type
@validate_email
async def create_community(steam_id: str, community_name: str, email: str,
                           disabled: bool = False, demos: bool = True,
                           community_type: str = None,
                           allow_api_access: bool = False,
                           match_start_webhook: str = None,
                           round_end_webhook: str = None,
                           match_end_webhook: str = None,
                           banned: bool = False
                           ) -> Tuple[CommunityModel, Community]:
    """Used to create community.

    Parameters
    ----------
    steam_id : str
    community_name : str
    email : str
    disabled : bool, optional
        by default False
    demos : bool, optional
        by default True
    community_type : str, optional
        by default None
    allow_api_access : bool, optional
        by default False
    match_start_webhook : str, optional
        by default None
    round_end_webhook : str, optional
        by default None
    match_end_webhook : str, optional
        by default None
    banned : bool, optional
        by default False

    Returns
    -------
    CommunityModel
    Community
        Used for interacting with a community

    Raises
    ------
    CommunityTaken
        Raised when community name is taken.
    AlreadyCommunity
        Raised when owner already owns a community.
    InvalidCommunityName
        Raised when community name isn't alphanumeric
        or character length is above 32 or below 4.
    InvalidCommunityType
        Raised when community type isn't valid.
    InvalidUploadSize
        Raised when upload size is incorrect.
    """

    if community_type:
        community_type_id = Config.community_types[community_type]
    else:
        community_type_id = None

    if await owner_exists(steam_id):
        raise AlreadyCommunity()

    try:
        await create_user(steam_id, "Unknown")
    except UserExists:
        pass

    customer, _ = await Sessions.stripe.create_customer(
        name=community_name
    )

    now = datetime.now()

    query = community_table.insert().values(
        community_name=community_name,
        owner_id=steam_id,
        disabled=disabled,
        demos=demos,
        timestamp=now,
        community_type_id=community_type_id,
        allow_api_access=allow_api_access,
        match_start_webhook=match_start_webhook,
        round_end_webhook=round_end_webhook,
        match_end_webhook=match_end_webhook,
        customer_id=customer.id,
        email=email,
        banned=banned
    )

    try:
        await Sessions.database.execute(query=query)
    except Exception:
        raise CommunityTaken()
    else:
        api_key = token_urlsafe(24)

        query = api_key_table.insert().values(
            api_key=api_key,
            owner_id=steam_id,
            timestamp=now,
            community_name=community_name,
            master=True
        )

        await Sessions.database.execute(query=query)

        return CommunityModel(
            api_key=api_key,
            owner_id=steam_id,
            disabled=disabled,
            community_name=community_name,
            timestamp=now,
            allow_api_access=allow_api_access,
            match_start_webhook=match_start_webhook,
            match_end_webhook=match_end_webhook,
            round_end_webhook=round_end_webhook,
            email=email,
            banned=banned
        ), Community(community_name)



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


from typing import Any
from .resources import Sessions


class CacheBase:
    def __init__(self, key: str) -> None:
        self.key = key

    async def expire(self) -> None:
        await Sessions.cache.delete(self.key)

    async def set(self, value: Any, ttl=None) -> None:
        await Sessions.cache.set(self.key, value, ttl=ttl)

    async def get(self) -> Any:
        return await Sessions.cache.get(self.key)


class __ScoreboardCache:
    def scoreboard(self, match_id: str) -> CacheBase:
        return CacheBase(self.key + "-" + match_id)


class __MatchesCache:
    def matches(self) -> CacheBase:
        return CacheBase(self.key + "-matches")


class __ProfileCache:
    def profile(self, steam_id: str) -> CacheBase:
        return CacheBase(self.key + steam_id)


class CommunityCache(CacheBase, __MatchesCache, __ScoreboardCache,
                     __ProfileCache):
    def __init__(self, community_name: str) -> None:
        super().__init__(community_name)

    def stats(self, key: str = "-stats") -> CacheBase:
        return CacheBase(self.key + key)

    def payments(self, key: str = "-payments") -> CacheBase:
        return CacheBase(self.key + key)


class CommunitiesCache(CacheBase, __MatchesCache):
    def __init__(self, key: str = "communities") -> None:
        super().__init__(key)


class VersionCache(CacheBase):
    def __init__(self, major: int, minor: int, patch: int) -> None:
        super().__init__("version-{}-{}-{}".format(major, minor, patch))


class VersionsCache(CacheBase):
    def __init__(self, key: str = "versions") -> None:
        super().__init__(key)


class ServerCache(CacheBase):
    def __init__(self, ip: str, port: int) -> None:
        super().__init__("{}-{}".format(ip, port))


class ServersCache(CacheBase):
    def __init__(self, community_name: str) -> None:
        super().__init__(community_name + "-servers")



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from typing import Any, Dict, Generator, List
from datetime import datetime

from ..resources import Config


class _DepthStatsModel:
    def __init__(self, kills: int, deaths: int,
                 headshots: int, shots_hit: int,
                 shots_fired: int) -> None:
        self.kills = kills
        self.deaths = deaths
        self.headshots = headshots
        self.shots_hit = shots_hit
        self.shots_fired = shots_fired

    @property
    def kdr(self) -> float:
        return (
            round(self.kills / self.deaths, 2)
            if self.kills > 0 and self.deaths > 0 else 0.00
        )

    @property
    def hs_percentage(self) -> float:
        return (
            round((self.headshots / self.kills) * 100, 2)
            if self.kills > 0 and self.headshots > 0 else 0.00
        )

    @property
    def hit_percentage(self) -> float:
        return (
            round((self.shots_hit / self.shots_fired) * 100, 2)
            if self.shots_fired > 0 and self.shots_hit > 0 else 0.00
        )


class PublicCommunityModel:
    def __init__(self, owner_id: str, disabled: bool, community_name: str,
                 timestamp: datetime, banned: bool,
                 allow_api_access: bool) -> None:
        self.owner_id = owner_id
        self.disabled = disabled
        self.community_name = community_name
        self.timestamp = timestamp
        self.banned = banned
        self.allow_api_access = allow_api_access

    @property
    def api_schema(self) -> dict:
        return {
            "community_name": self.community_name,
            "owner_id": self.owner_id,
            "disabled": self.disabled,
            "timestamp": self.timestamp.strftime(Config.timestamp_format),
            "banned": self.banned,
            "allow_api_access": self.allow_api_access
        }


class CommunityModel(PublicCommunityModel):
    def __init__(self, api_key: str,
                 match_start_webhook: str,
                 round_end_webhook: str,
                 match_end_webhook: str,
                 email: str,
                 amount: float = None,
                 customer_id: str = None,
                 cancelled: bool = None,
                 subscription_expires: datetime = None,
                 **kwargs) -> None:
        super().__init__(**kwargs)

        self.master_api_key = api_key
        self.amount = amount if amount else 0.0
        self.match_start_webhook = match_start_webhook
        self.round_end_webhook = round_end_webhook
        self.match_end_webhook = match_end_webhook
        self.customer_id = customer_id
        self.email = email
        self.cancelled = cancelled
        self.subscription_expires = subscription_expires

    @property
    def api_schema(self) -> dict:
        return {
            "master_api_key": self.master_api_key,
            "amount": self.amount,
            "match_start_webhook": self.match_start_webhook,
            "round_end_webhook": self.round_end_webhook,
            "match_end_webhook": self.match_end_webhook,
            "customer_id": self.customer_id,
            "email": self.email,
            "cancelled": self.cancelled,
            "subscription_expires": self.subscription_expires.strftime(
                Config.timestamp_format
            ) if self.subscription_expires else None,
            **super().api_schema
        }


class MatchModel:
    def __init__(self, match_id: str, timestamp: datetime, status: int,
                 demo_status: int, map: str, team_1_name: str,
                 team_2_name: str, team_1_score: int,
                 team_2_score: int, team_1_side: int,
                 team_2_side: int, community_name: str) -> None:
        self.match_id = match_id
        self.timestamp = timestamp
        self.status = status
        self.demo_status = demo_status
        self.map = map
        self.team_1_name = team_1_name
        self.team_2_name = team_2_name
        self.team_1_score = team_1_score
        self.team_2_score = team_2_score
        self.team_1_side = team_1_side
        self.team_2_side = team_2_side
        self.cover_image = "{}maps/{}".format(
            Config.url,
            Config.map_images[self.map] if self.map in Config.map_images
            else "invalid.png"
        )
        self.community_name = community_name

    @property
    def api_schema(self) -> dict:
        return {
            "match_id": self.match_id,
            "timestamp": self.timestamp.strftime(Config.timestamp_format),
            "status": self.status,
            "demo_status": self.demo_status,
            "map": self.map,
            "team_1_name": self.team_1_name,
            "team_2_name": self.team_2_name,
            "team_1_score": self.team_1_score,
            "team_2_score": self.team_2_score,
            "team_1_side": self.team_1_side,
            "team_2_side": self.team_2_side,
            "cover_image": self.cover_image,
            "community_name": self.community_name
        }


class ProfileOverviewModel:
    def __init__(self, name: str, steam_id: str, kills: int, headshots: int,
                 assists: int, deaths: int) -> None:
        self.name = name
        self.steam_id = steam_id
        self.kills = kills
        self.headshots = headshots
        self.assists = assists
        self.deaths = deaths

    @property
    def api_schema(self) -> dict:
        return {
            "name": self.name,
            "steam_id": self.steam_id,
            "kills": self.kills,
            "headshots": self.headshots,
            "assists": self.assists,
            "deaths": self.deaths,
        }


class ProfileModel(ProfileOverviewModel, _DepthStatsModel):
    def __init__(self, shots_fired: int, shots_hit: int,
                 mvps: int, timestamp: datetime, **kwargs) -> None:
        super().__init__(**kwargs)

        _DepthStatsModel.__init__(
            self, kwargs["kills"], kwargs["deaths"],
            kwargs["headshots"], shots_hit, shots_fired
        )

        self.shots_fired = shots_fired
        self.shots_hit = shots_hit
        self.mvps = mvps
        self.timestamp = timestamp

    @property
    def api_schema(self) -> dict:
        return {
            **super().api_schema,
            "kdr": self.kdr,
            "hs_percentage": self.hs_percentage,
            "hit_percentage": self.hit_percentage,
            "shots_fired": self.shots_fired,
            "shots_hit": self.shots_hit,
            "mvps": self.mvps,
            "timestamp": self.timestamp.strftime(Config.timestamp_format)
        }


class _ScoreboardPlayerModel(_DepthStatsModel):
    def __init__(self, name: str, steam_id: str, team: int,
                 alive: bool, ping: int, kills: int, headshots: int,
                 assists: int, deaths: int, shots_fired: int,
                 shots_hit: int, mvps: int, score: int,
                 disconnected: bool) -> None:
        super().__init__(kills, deaths,
                         headshots, shots_hit, shots_fired)

        self.name = name
        self.steam_id = steam_id
        self.team = team
        self.alive = alive
        self.ping = ping
        self.kills = kills
        self.headshots = headshots
        self.assists = assists
        self.deaths = deaths
        self.shots_fired = shots_fired
        self.shots_hit = shots_hit
        self.mvps = mvps
        self.score = score
        self.disconnected = disconnected


class ScoreboardModel(MatchModel):
    def __init__(self, team_1: List[Dict[str, Any]],
                 team_2: List[Dict[str, Any]], match: Dict[str, Any]) -> None:
        super().__init__(**match)

        self.__team_1 = team_1
        self.__team_2 = team_2

    def team_1(self) -> Generator[_ScoreboardPlayerModel, None, None]:
        """Lists players in team 1.

        Yields
        ------
        _ScoreboardPlayerModel
            Holds player data.
        """

        for player in self.__team_1:
            yield _ScoreboardPlayerModel(**player)

    def team_2(self) -> Generator[_ScoreboardPlayerModel, None, None]:
        """Lists players in team 2.

        Yields
        ------
        _ScoreboardPlayerModel
            Holds player data.
        """

        for player in self.__team_2:
            yield _ScoreboardPlayerModel(**player)

    @property
    def api_schema(self) -> dict:
        return {
            **super().api_schema,
            "team_1": self.__team_1,
            "team_2": self.__team_2
        }


class CommunityStatsModel:
    def __init__(self, total_matches: int, active_matches: int,
                 stored_demos: int, total_users: int) -> None:
        self.total_matches = total_matches
        self.active_matches = active_matches
        self.stored_demos = stored_demos
        self.total_users = total_users

    @property
    def api_schema(self) -> dict:
        return {
            "total_matches": self.total_matches,
            "active_matches": self.active_matches,
            "stored_demos": self.stored_demos,
            "total_users": self.total_users
        }


class ServerModel:
    def __init__(self, community_name: str, ip: str,
                 port: int, name: str, players: int,
                 max_players: int, map: str) -> None:
        self.community_name = community_name
        self.ip = ip
        self.port = port
        self.name = name
        self.players = players
        self.max_players = max_players
        self.map = map
        self.cover_image = "{}maps/{}".format(
            Config.url,
            Config.map_images[self.map] if self.map in Config.map_images
            else "invalid.png"
        )

    @property
    def api_schema(self) -> dict:
        return {
            "community_name": self.community_name,
            "name": self.name,
            "players": self.players,
            "max_players": self.max_players,
            "ip": self.ip,
            "port": self.port,
            "map": self.map,
            "cover_image": self.cover_image
        }



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


class SQLMatchesException(Exception):
    """Base Exception for SQLMatches.
    """

    def __init__(self, msg="SQLMatches Exception", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class UnSupportedEngine(SQLMatchesException):
    """Raised when the database engine isn't supported.
    """

    def __init__(self, msg="SQL Engine not supported", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class CommunityTaken(SQLMatchesException):
    """Raised when community name is taken.
    """

    def __init__(self, msg="Community name taken", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class AlreadyCommunity(SQLMatchesException):
    """Raised when user already owns a community.
    """

    def __init__(self, msg="User already owns a community", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class InvalidCommunity(SQLMatchesException):
    """Raised when community ID doesn't exist.
    """

    def __init__(self, msg="Invalid Community ID", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class NoOwnership(SQLMatchesException):
    """Raised when steam id doesn't own any communties.
    """

    def __init__(self, msg="User owns no communities", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class InvalidMatchID(SQLMatchesException):
    """Raised when match ID is invalid.
    """

    def __init__(self, msg="Invalid Match ID", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class InvalidAPIKey(SQLMatchesException):
    """Raised when API key is invalid.
    """

    def __init__(self, msg="Invalid API Key", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class DemoAlreadyUploaded(SQLMatchesException):
    """Raised when a demo has already been uploaded.
    """

    def __init__(self, msg="Demo already uploaded", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class InvalidSteamID(SQLMatchesException):
    """Raised when Steam ID isn't valid
    """

    def __init__(self, msg="Invalid Steam ID", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class InvalidCommunityName(SQLMatchesException):
    """Raised when community name isn't alphanumeric
       or character length is above 32 or below 4.
    """

    def __init__(self, msg="Commany Name not alphanumeric", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class InvalidCommunityType(SQLMatchesException):
    """Raised when community type isn't valid.
    """

    def __init__(self, msg="Commany type invalid", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class UserExists(SQLMatchesException):
    """Raised when user exists.
    """

    def __init__(self, msg="User exists", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class InvalidWebhook(SQLMatchesException):
    """Raised when webhook URL is invalid.
    """

    def __init__(self, msg="Invalid webhook url", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class InvalidEmail(SQLMatchesException):
    """Raised when email is invalid.
    """

    def __init__(self, msg="Invalid email", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class InvalidCustomer(SQLMatchesException):
    """Raised when customer ID is invalid.
    """

    def __init__(self, msg="Invalid stripe customer ID", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class InvalidVersion(SQLMatchesException):
    """Raised when version invalid.
    """

    def __init__(self, msg="Invalid version", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class ServerExists(SQLMatchesException):
    """Raised when server already exists.
    """

    def __init__(self, msg="Server exists", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class InvalidServer(SQLMatchesException):
    """Raised when server doesn't exists.
    """

    def __init__(self, msg="Invalid server", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


import socketio

from starlette.routing import Route, Mount
from starlette.staticfiles import StaticFiles
from starlette.exceptions import HTTPException

from webargs_starlette import WebargsHTTPException

from ..exceptions import SQLMatchesException
from ..resources import Config, Sessions

# Routes
from .api.matches import (
    MatchAPI,
    CreateMatchAPI,
    DemoUploadAPI,
    MatchesAPI
)
from .api.players import CommunityPlayersAPI
from .api.misc import SchemaAPI
from .api.community import (
    CommunityOwnerAPI,
    CommunityCreateAPI,
    CommunityOwnerMatchesAPI,
    CommunityUpdateAPI,
    CommunityExistsAPI,
    PublicCommunityAPI,
    CommunitySessionAPI
)
from .api.key import KeyAPI
from .api.communities import (
    CommunitiesAPI,
    CommunityMatchesAPI,
    MatchesCommunitiesAPI
)
from .api.admin import (
    CommunitiesAdminAPI,
    AdminAPI,
    SavePluginAPI
)
from .api.version import VersionAPI, VersionsAPI
from .api.profile import ProfileAPI, SteamProfileCors
from .api.server import ServerAPI, ServersAPI
from .api.auto_setup import AutoSetupAPI

# A bit gross, but because socketio uses singletons, we
# need to do this.
from .websockets import *  # noqa: F403, F401

from .webhooks import (
    PaymentFailedWebhook,
    PaymentSuccessWebhook
)

from .download import DownloadPage
from .steam import (
    SteamValidate,
    SteamLogin,
    SteamLogout
)
from .errors import (
    server_error,
    payload_error,
    internal_error
)


ERROR_HANDLERS = {
    WebargsHTTPException: payload_error,
    HTTPException: server_error,
    SQLMatchesException: internal_error
}


ROUTES = [
    Mount("/api", name="api", routes=[
        Mount("/steam", routes=[
            Route("/login", SteamLogin),
            Route("/validate", SteamValidate),
            Route("/logout", SteamLogout)
        ]),
        Mount("/maps/", StaticFiles(directory=Config.maps_dir), name="maps"),
        Route("/matches/", MatchesAPI),  # Tested - POST @ 0.2.0
        Mount("/match", routes=[
            Route("/create/", CreateMatchAPI),  # Tested - POST @ 0.2.0
            Mount("/{match_id}", routes=[
                Route("/", MatchAPI),  # Tested - GET, POST, DELETE @ 0.2.0
                Route("/upload/", DemoUploadAPI),
                Route("/download/", DownloadPage, name="DownloadPage")
            ])
        ]),
        Route("/players/", CommunityPlayersAPI),
        Mount("/profile/{steam_id}", routes=[
            Route("/cros/", SteamProfileCors),
            Route("/", ProfileAPI)
        ]),
        Mount("/version", routes=[
            Route("/{major:int}/{minor:int}/{patch:int}/", VersionAPI),
            Route("/", VersionsAPI)
        ]),
        Route("/servers/", ServersAPI),
        Route("/server/{ip:str}/{port:int}/", ServerAPI),
        Mount("/community", routes=[
            Route("/exists/", CommunityExistsAPI),
            Mount("/owner", routes=[
                Route("/", CommunityOwnerAPI),
                Route("/matches/", CommunityOwnerMatchesAPI),
                Route("/update/", CommunityUpdateAPI),
                Route("/stripe-session/", CommunitySessionAPI),
                Route("/autosetup/", AutoSetupAPI)
            ]),
            Route("/key/", KeyAPI),
            Route("/public/", PublicCommunityAPI),
            Route("/", CommunityCreateAPI),
        ]),
        Mount("/communities", routes=[
            Route("/", CommunitiesAPI),
            Route("/matches/", CommunityMatchesAPI),
            Route("/all/", MatchesCommunitiesAPI)
        ]),
        Mount("/admin", routes=[
            Route("/communities/", CommunitiesAdminAPI),
            Route("/plugins/", SavePluginAPI),
            Route("/", AdminAPI)
        ]),
        Route("/schema/", SchemaAPI, include_in_schema=False)
    ]),
    Mount("/ws/", socketio.ASGIApp(Sessions.websocket)),
    Mount("/webhook", routes=[
        Mount("/payment", routes=[
            Route("/fail/", PaymentFailedWebhook),
            Route("/success/", PaymentSuccessWebhook)
        ])
    ])
]



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


import logging
import bcrypt
import backblaze
import aioftp

from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.sessions import SessionMiddleware
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.middleware.cors import CORSMiddleware
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles

from typing import Dict, List, Tuple

from datetime import timedelta

from databases import Database
from aiohttp import ClientSession
from aiojobs import create_scheduler
from aiocache import Cache
from aiosmtplib import SMTP

from .stripe import Stripe

from .tables import create_tables
from .resources import Sessions, Config
from .settings import (
    DatabaseSettings,
    B2UploadSettings,
    LocalUploadSettings,
    StripeSettings,
    SmtpSettings,
    WebhookSettings
)
from .middlewares import APIAuthentication

from .routes import ROUTES, ERROR_HANDLERS
from .routes.errors import auth_error

from .background_tasks import TASKS_TO_SPAWN

from .misc import cache_community_types

from .key_loader import KeyLoader

from .constants import MAP_IMAGES, COMMUNITY_TYPES


__version__ = "0.3.6"
__url__ = "https://github.com/WardPearce/SQLMatches"
__description__ = "SQLMatches, match & demos recorder."
__author__ = "WardPearce"
__author_email__ = "wardpearce@protonmail.com"
__license__ = "GPL v3"


logger = logging.getLogger("SQLMatches")


class SQLMatches(Starlette):
    def __init__(self,
                 database_settings: DatabaseSettings,
                 stripe_settings: StripeSettings,
                 smtp_settings: SmtpSettings,
                 friendly_url: str,
                 frontend_url: str,
                 root_steam_id: str,
                 system_email: str,
                 upload_settings: Tuple[
                     B2UploadSettings, LocalUploadSettings] = None,
                 map_images: Dict[str, str] = MAP_IMAGES,
                 upload_delay: float = 0.00001,
                 free_upload_size: float = 30.0,
                 max_upload_size: float = 100.0,
                 timestamp_format: str = "%m/%d/%Y-%H:%M:%S",
                 community_types: List[str] = COMMUNITY_TYPES,
                 webhook_settings: WebhookSettings = WebhookSettings(),
                 match_max_length: timedelta = timedelta(hours=3),
                 demo_expires: timedelta = timedelta(weeks=20),
                 subscription_length: timedelta = timedelta(days=31),
                 clear_cache: bool = True,
                 **kwargs) -> None:
        """SQLMatches API.

        Parameters
        ----------
        database_settings : DatabaseSettings
        stripe_settings : StripeSettings
        friendly_url : str
        frontend_url : str
        root_steam_id : str
        system_email : str
        upload_settings : [B2UploadSettings, LocalUploadSettings], optional
            by default None
        map_images : Dict[str, str], optional
            by default MAP_IMAGES
        upload_delay : float, optional
            by default 0.00001
        free_upload_size : float, optional
            by default 50.0
        max_upload_size : float, optional
            by default 100.0
        timestamp_format : str, optional
            by default "%m/%d/%Y-%H:%M:%S"
        community_types : List[str], optional
            by default COMMUNITY_TYPES
        webhook_settings : WebhookSettings, optional
            by default WebhookSettings()
        match_max_length : timedelta, optional
            by default timedelta(hours=3)
        clear_cache : bool, optional
            by default True
        demo_expires : timedelta, optional
            by default timedelta(weeks=20)
        """

        startup_tasks = [self._startup]
        shutdown_tasks = [self._shutdown]

        if "on_startup" in kwargs:
            startup_tasks = startup_tasks + kwargs["on_startup"]

        if "on_shutdown" in kwargs:
            shutdown_tasks = shutdown_tasks + kwargs["on_shutdown"]

        middlewares = [
            Middleware(SessionMiddleware,
                       secret_key=KeyLoader(name="session").load()),
            Middleware(AuthenticationMiddleware, backend=APIAuthentication(),
                       on_error=auth_error),
            Middleware(
                CORSMiddleware,
                allow_origins=["*"],
                allow_methods=["GET", "POST", "DELETE", "OPTIONS"]
            )
        ]

        if "middleware" in kwargs:
            middlewares = middlewares + kwargs["middleware"]

        if "routes" in kwargs:
            routes = kwargs["routes"] + ROUTES
        else:
            routes = ROUTES

        if "exception_handlers" in kwargs:
            exception_handlers = kwargs["exception_handlers"] + ERROR_HANDLERS
        else:
            exception_handlers = ERROR_HANDLERS

        if friendly_url[:1] != "/":
            friendly_url += "/"

        if frontend_url[:1] != "/":
            frontend_url += "/"

        Config.url = friendly_url
        Config.map_images = map_images
        Config.upload_delay = upload_delay
        Config.free_upload_size = free_upload_size
        Config.max_upload_size = max_upload_size
        Config.timestamp_format = timestamp_format
        Config.root_steam_id_hashed = bcrypt.hashpw(
            root_steam_id.encode(), bcrypt.gensalt()
        )
        Config.root_webhook_key_hashed = bcrypt.hashpw(
            (KeyLoader("webhook").load()).encode(), bcrypt.gensalt()
        )

        Config.webhook_timeout = webhook_settings.timeout
        Config.webhook_match_end = webhook_settings.match_end
        Config.webhook_match_start = webhook_settings.match_start
        Config.webhook_round_end = webhook_settings.round_end
        Config.webhook_key = webhook_settings.key

        Config.match_max_length = match_max_length
        Config.system_email = system_email
        Config.frontend_url = frontend_url
        Config.demo_expires = demo_expires
        Config.price_id = stripe_settings.price_id
        Config.subscription_length = subscription_length

        self.community_types = community_types
        self.clear_cache = clear_cache

        database_url = "://{}:{}@{}:{}/{}?charset=utf8mb4".format(
            database_settings.username,
            database_settings.password,
            database_settings.server,
            database_settings.port,
            database_settings.database
        )

        Sessions.database = Database(
            database_settings.engine + database_url
        )

        create_tables(
            "{}+{}{}".format(
                database_settings.engine,
                database_settings.alchemy_engine,
                database_url
            )
        )

        Sessions.smtp = SMTP(
            hostname=smtp_settings.hostname,
            port=smtp_settings.port,
            use_tls=smtp_settings.use_tls,
            password=smtp_settings.password,
            username=smtp_settings.username
        )

        Sessions.stripe = Stripe(
            stripe_settings.api_key,
            stripe_settings.testing
        )

        Sessions.ftp = aioftp.Client()

        if upload_settings:
            Config.demo_pathway = upload_settings.pathway
            Config.demo_extension = upload_settings.extension

            if isinstance(upload_settings, B2UploadSettings):
                Config.cdn_url = upload_settings.cdn_url
                Config.upload_type = B2UploadSettings

                self.b2 = backblaze.Awaiting(
                    upload_settings.key_id,
                    upload_settings.application_key
                )

                Sessions.bucket = self.b2.bucket(
                    upload_settings.bucket_id
                )

            elif isinstance(upload_settings, LocalUploadSettings):
                Config.cdn_url = None
                Config.upload_type = LocalUploadSettings

                # Dynamically adding mount if local storage.

                # Name attribute for Mount isn't working correctly here,
                # so please don't change '/demos/'.
                for mount in routes:
                    if type(mount) == Mount and mount.name == "api":
                        mount.app.routes.append(
                            Mount(
                                "/demos/",
                                StaticFiles(directory=Config.demo_pathway)
                            )
                        )
                        break

                logger.warning(
                    "Using local storage for demos, use b2 for production."
                )
        else:
            Config.upload_type = None

        super().__init__(
            routes=routes,
            exception_handlers=exception_handlers,
            middleware=middlewares,
            on_startup=startup_tasks,
            on_shutdown=shutdown_tasks,
            **kwargs
        )

    async def _startup(self) -> None:
        """Creates needed sessions.
        """

        await Sessions.smtp.connect()
        await Sessions.database.connect()
        Sessions.aiohttp = ClientSession()

        try:
            Sessions.cache = Cache(Cache.REDIS)
            await Sessions.cache.exists("connection")
        except ConnectionRefusedError:
            Sessions.cache = Cache(Cache.MEMORY)
            logger.warning(
                "Memory cache being used, use redis for production."
            )

        if self.clear_cache:
            await Sessions.cache.clear()

        if Config.upload_type == B2UploadSettings:
            await self.b2.authorize()

        self.background_tasks = await create_scheduler()
        for to_spawn in TASKS_TO_SPAWN:
            await self.background_tasks.spawn(to_spawn())

        await cache_community_types(self.community_types)

    async def _shutdown(self) -> None:
        """Closes any underlying sessions.
        """

        await Sessions.smtp.quit()
        await Sessions.database.disconnect()
        await Sessions.aiohttp.close()
        await Sessions.cache.close()

        if Config.upload_type == B2UploadSettings:
            await self.b2.close()

        await self.background_tasks.close()



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from asyncio import sleep
import logging
from sqlalchemy.sql import select, and_, or_, func, text
from datetime import datetime

from .demos import Demo
from .resources import DemoQueue, Sessions, Config
from .tables import scoreboard_total_table
from .caches import CommunityCache
from .community.match import Match


async def demo_delete() -> None:
    """Handles deleting demos in background off local storage or B2.
    """

    while True:
        for community_name, matches in dict(DemoQueue.matches).items():
            for match_id in matches:
                demo = Demo(Match(
                    match_id,
                    community_name
                ))

                if demo.delete:
                    await demo.delete()
                else:
                    break

                await sleep(5.0)

            DemoQueue.matches.pop(community_name, None)

        await sleep(10.0)


async def expired_demos() -> None:
    """Used to expire demos what are older then the max
    life-span for a demo.
    """

    while True:
        query = select([
            scoreboard_total_table.c.match_id,
            scoreboard_total_table.c.community_name
        ]).select_from(scoreboard_total_table).where(
            and_(
                datetime.now() >= func.timestampadd(
                    text("DAY"),
                    Config.demo_expires.days,
                    scoreboard_total_table.c.timestamp
                ),
                scoreboard_total_table.c.demo_status == 2
            )
        )

        statements = []
        statements_append = statements.append
        async for match in Sessions.database.iterate(query):
            logging.info("Attempting to delete demo of {}".format(
                match["match_id"]
            ))

            if match["community_name"] not in DemoQueue.matches:
                DemoQueue.matches[match["community_name"]] = []

            DemoQueue.matches[match["community_name"]].append(
                match["match_id"]
            )

            statements_append(
                and_(
                    scoreboard_total_table.c.match_id == match["match_id"],
                    scoreboard_total_table.c.community_name ==
                    match["community_name"]
                )
            )

            await (CommunityCache(
                match["community_name"]
            ).scoreboard(match["match_id"])).expire()

        if statements:
            await Sessions.database.execute(
                scoreboard_total_table.update().values(
                    demo_status=4
                ).where(
                    or_(
                        *statements
                    )
                )
            )

        await sleep(43200.0)


async def match_ender() -> None:
    """If a match doesn't get ended correctly, this will
    close it after X amount of time.
    """

    # I'm sorry David, please forgive me. Maybe just turn the big telly on.
    while True:
        query = select([
            scoreboard_total_table.c.match_id,
            scoreboard_total_table.c.community_name
        ]).select_from(
            scoreboard_total_table
        ).where(
            and_(
                scoreboard_total_table.c.status == 1,
                datetime.now() + Config.match_max_length >=
                scoreboard_total_table.c.timestamp
            )
        )

        statements = []
        statements_append = statements.append
        async for match in Sessions.database.iterate(query):
            logging.info("Attempting to end match {}".format(
                match["match_id"]
            ))

            statements_append(
                and_(
                    scoreboard_total_table.c.match_id == match["match_id"],
                    scoreboard_total_table.c.community_name ==
                    match["community_name"]
                )
            )

            await (CommunityCache(
                match["community_name"]
            ).scoreboard(match["match_id"])).expire()

        if statements:
            await Sessions.database.execute(
                scoreboard_total_table.update().values(
                    status=0
                ).where(
                    or_(
                        *statements
                    )
                )
            )

        await sleep(400.0)


TASKS_TO_SPAWN = [
    demo_delete,
    match_ender,
    expired_demos
]



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


import binascii
import bcrypt
from base64 import b64decode

from typing import Tuple

from starlette.authentication import (
    AuthenticationBackend,
    AuthenticationError,
    SimpleUser,
    AuthCredentials
)
from starlette.requests import Request

from .community import (
    Community,
    api_key_to_community,
    get_community_from_owner
)
from .resources import Config
from .exceptions import InvalidAPIKey, NoOwnership


AUTH_ERROR = "Invalid basic auth credentials"


class APIAuthentication(AuthenticationBackend):
    async def authenticate(self, request: Request
                           ) -> Tuple[AuthCredentials, SimpleUser]:
        """Used to authenticate basic auth.

        Parameters
        ----------
        request : Request

        Returns
        -------
        AuthCredentials
        SimpleUser

        Raises
        ------
        AuthenticationError
        """

        if "Authorization" in request.headers:
            auth = request.headers["Authorization"]
            try:
                scheme, credentials = auth.split()
                if scheme.lower() != "basic":
                    return
                decoded = b64decode(credentials).decode("ascii")
            except (ValueError, UnicodeDecodeError, binascii.Error):
                raise AuthenticationError(AUTH_ERROR)

            username, _, password = decoded.partition(":")

            try:
                request.state.community, master = await api_key_to_community(
                    password
                )
            except InvalidAPIKey:
                raise AuthenticationError(AUTH_ERROR)
            else:
                return AuthCredentials([
                    "community", "master" if master else None
                ]), SimpleUser(username)

        elif "steam_id" in request.session:
            scopes = ["steam_login"]

            if "community_name" in request.query_params:
                scopes.append("community")

                if ("check_ownership" in request.query_params and
                        request.query_params["check_ownership"].lower()
                        == "true"):
                    try:
                        (
                            community,
                            banned,
                            active_subscription
                        ) = await get_community_from_owner(
                            request.session["steam_id"])

                        if banned:
                            return

                    except NoOwnership:
                        pass
                    else:
                        if (community.community_name ==
                                request.query_params["community_name"]):
                            scopes.append("is_owner")

                            if ("check_subscription" in request.query_params
                                and request.query_params["check_subscription"]
                                .lower() == "true" and
                                    active_subscription):

                                scopes.append("active_subscription")

                request.state.community = Community(
                    request.query_params["community_name"]
                )

            if("check_root" in request.query_params and
               request.query_params["check_root"].lower() == "true" and
               bcrypt.checkpw(request.session["steam_id"].encode(),
                              Config.root_steam_id_hashed)):

                scopes.append("root_login")

            return (
                AuthCredentials(scopes),
                SimpleUser(request.session["steam_id"])
            )

        elif "webhook_key" in request.query_params:
            if bcrypt.checkpw(request.query_params["webhook_key"].encode(),
                              Config.root_webhook_key_hashed):
                return (
                    AuthCredentials(["stripe_webhook"]),
                    SimpleUser("")
                )



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


import os
import socketio
import aioftp

from typing import Any, Dict

from backblaze.bucket.awaiting import AwaitingBucket
from aiohttp import ClientSession
from databases import Database
from aiocache import Cache
from datetime import timedelta
from aiosmtplib import SMTP


class Sessions:
    database: Database
    aiohttp: ClientSession
    bucket: AwaitingBucket
    cache: Cache
    websocket = socketio.AsyncServer(
        async_mode="asgi",
        cors_allowed_origins=[]
    )
    stripe: Any
    smtp: SMTP
    ftp: aioftp.Client


class Config:
    current_dir = os.path.dirname(os.path.realpath(__file__))
    maps_dir = os.path.join(
        current_dir,
        "maps"
    )
    plugin_dir = os.path.join(
        current_dir,
        "plugins"
    )
    steam_openid_url = "https://steamcommunity.com/openid/login"
    upload_type: Any
    url: str
    map_images: str
    db_engine: str
    demo_extension: str
    cdn_url: str
    demo_pathway: str
    upload_delay: float
    free_upload_size: float
    max_upload_size: float
    timestamp_format: str
    root_steam_id_hashed: Any
    root_webhook_key_hashed: Any
    # Type string, type ID
    community_types: Dict[str, int] = {}
    webhook_timeout: int
    webhook_match_start: str
    webhook_match_end: str
    webhook_round_end: str
    webhook_key: str
    match_max_length: timedelta
    demo_expires: timedelta
    subscription_length: timedelta
    system_email: str
    frontend_url: str
    price_id: str


class DemoQueue:
    matches: dict = {}



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from sqlalchemy.sql import and_

from ..resources import Sessions
from ..tables import server_table
from ..exceptions import InvalidServer

from .models import ServerModel


class Server:
    def __init__(self, ip: str, port: int, community_name: str) -> None:
        self.ip = ip
        self.port = port
        self.community_name = community_name

    @property
    def __and_statement(self) -> and_:
        return and_(
            server_table.c.ip == self.ip,
            server_table.c.port == self.port,
            server_table.c.community_name == self.community_name
        )

    async def get(self) -> ServerModel:
        """Used to get server.

        Returns
        -------
        ServerModel

        Raises
        ------
        InvalidServer
        """

        row = await Sessions.database.fetch_one(
            server_table.select().where(
                self.__and_statement
            )
        )

        if row:
            return ServerModel(**row)
        else:
            raise InvalidServer()

    async def delete(self) -> None:
        """Used to delete server.
        """

        await Sessions.database.execute(
            server_table.delete().where(
                self.__and_statement
            )
        )

    async def update(self, players: int = None,
                     max_players: int = None,
                     ip: str = None, port: int = None,
                     name: str = None, map_name: str = None) -> None:
        """Used to update server.

        Parameters
        ----------
        players : int, optional
            by default None
        max_players : int, optional
            by default None
        ip : str, optional
            by default None
        port : str, optional
            by default None
        name : str, optional
            by default None
        map_name : str, optional
            by default None
        """

        values = {}
        if players is not None:
            values["players"] = players
        if max_players is not None:
            values["max_players"] = max_players
        if port is not None:
            values["port"] = port
        if ip:
            values["ip"] = ip
        if name:
            values["name"] = name
        if map_name:
            values["map"] = map_name

        if values:
            await Sessions.database.execute(
                server_table.update().values(
                    **values
                ).where(self.__and_statement)
            )



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


from starlette.endpoints import HTTPEndpoint
from starlette.authentication import requires
from starlette.requests import Request
from starlette.background import BackgroundTask

from marshmallow import Schema, validate
from webargs import fields
from webargs_starlette import use_args

from ...webhook_pusher import WebhookPusher
from ...responses import response
from ...resources import Sessions, Config
from ...demos import Demo
from ...caches import CommunityCache, CommunitiesCache
from ...exceptions import InvalidMatchID, DemoAlreadyUploaded


class PlayersSchema(Schema):
    name = fields.Str(min=1, max=42, required=True)
    steam_id = fields.Str(min=64, max=64)
    team = fields.Int(required=True)
    alive = fields.Bool(required=True)
    ping = fields.Int(required=True)
    kills = fields.Int(required=True)
    headshots = fields.Int(required=True)
    assists = fields.Int(required=True)
    deaths = fields.Int(required=True)
    shots_fired = fields.Int(required=True)
    shots_hit = fields.Int(required=True)
    mvps = fields.Int(required=True)
    score = fields.Int(required=True)
    disconnected = fields.Bool(required=True)


class MatchAPI(HTTPEndpoint):
    @requires("community")
    async def get(self, request: Request) -> response:
        """Used to get the scoreboard for a match.

        Parameters
        ----------
        request : Request
        """

        cache = CommunityCache(
            request.state.community.community_name
        ).scoreboard(request.path_params["match_id"])

        cache_get = await cache.get()
        if cache_get:
            return response(cache_get)

        try:
            scoreboard = await request.state.community.match(
                request.path_params["match_id"]
            ).scoreboard()
        except InvalidMatchID:
            raise
        else:
            data = scoreboard.api_schema

            await cache.set(data)

            return response(data)

    @use_args({"team_1_score": fields.Int(required=True,
                                          validates=validate.Range(0, 240)),
               "team_2_score": fields.Int(required=True,
                                          validates=validate.Range(0, 240)),
               "players": fields.List(fields.Nested(PlayersSchema),
                                      validates=validate.Length(1, 30)
                                      ),
               "team_1_side": fields.Int(validates=validate.Range(0, 1)),
               "team_2_side": fields.Int(validates=validate.Range(0, 1)),
               "end": fields.Bool()})
    @requires("master")
    async def post(self, request: Request, parameters: dict) -> response:
        """Used to update a match.

        Parameters
        ----------
        request : Request
        parameters : dict
        """

        match = request.state.community.match(
            request.path_params["match_id"]
        )

        try:
            await match.update(**parameters)
        except InvalidMatchID:
            raise
        else:
            scoreboard = await match.scoreboard()
            data = scoreboard.api_schema

            await ((CommunitiesCache()).matches()).expire()

            cache = CommunityCache(request.state.community.community_name)
            await (cache.matches()).expire()
            await (cache.scoreboard(request.path_params["match_id"])).set(
                data
            )

            await Sessions.websocket.emit(
                "match_update",
                data,
                room="ws_room"
            )

            await Sessions.websocket.emit(
                request.path_params["match_id"],
                data,
                room="ws_room"
            )

            pusher = WebhookPusher(
                request.state.community.community_name,
                data
            )

            return response(
                background=BackgroundTask(
                    pusher.match_end if "end" in parameters
                    and parameters["end"]
                    else pusher.round_end
                )
            )

    @requires("master")
    async def delete(self, request: Request) -> response:
        """Used to end a match.

        Parameters
        ----------
        request : Request
        """

        match = request.state.community.match(
            request.path_params["match_id"]
        )

        try:
            await match.end()
        except InvalidMatchID:
            raise
        else:
            try:
                scoreboard = await match.scoreboard()
            except InvalidMatchID:
                pass
            else:
                data = scoreboard.api_schema

                await ((CommunitiesCache()).matches()).expire()

                cache = CommunityCache(request.state.community.community_name)
                await (cache.matches()).expire()
                await (cache.scoreboard(request.path_params["match_id"])).set(
                    data
                )

                await Sessions.websocket.emit(
                    "match_update",
                    data,
                    room="ws_room"
                )

                await Sessions.websocket.emit(
                    request.path_params["match_id"],
                    data,
                    room="ws_room"
                )

                return response(
                    background=BackgroundTask(
                        WebhookPusher(
                            request.state.community.community_name,
                            data
                        ).match_end
                    )
                )

            return response()


class MatchesAPI(HTTPEndpoint):
    @use_args({"search": fields.Str(), "page": fields.Int(),
               "desc": fields.Bool(), "require_scoreboard": fields.Bool()})
    @requires("community")
    async def post(self, request: Request, parameters: dict) -> response:
        """Used to list matches.

        Parameters
        ----------
        request : Request
        parameters : dict
        """

        if not parameters:
            cache = CommunityCache(
                request.state.community.community_name
            ).matches()

            cache_get = await cache.get()
            if cache_get:
                return response(cache_get)

            data = [
                match.api_schema async for match, _ in
                request.state.community.matches(**parameters)
            ]

            await cache.set(data)
        else:
            data = [
                match.api_schema async for match, _ in
                request.state.community.matches(**parameters)
            ]

        return response(data)


class CreateMatchAPI(HTTPEndpoint):
    @use_args({"team_1_name": fields.Str(min=1, max=64, required=True),
               "team_2_name": fields.Str(min=1, max=64, required=True),
               "team_1_side": fields.Int(required=True,
                                         validates=validate.Range(0, 1)),
               "team_2_side": fields.Int(required=True,
                                         validates=validate.Range(0, 1)),
               "team_1_score": fields.Int(required=True,
                                          validates=validate.Range(0, 240)),
               "team_2_score": fields.Int(required=True,
                                          validates=validate.Range(0, 240)),
               "map_name": fields.Str(min=1, max=24, required=True)})
    @requires("master")
    async def post(self, request: Request, parameters: dict) -> response:
        """Used to create a match.

        Parameters
        ----------
        request : Request
        parameters : dict
        """

        data, match = await request.state.community.create_match(**parameters)

        await (CommunityCache(
            request.state.community.community_name
        ).matches()).expire()

        return response(
            {"match_id": match.match_id},
            background=BackgroundTask(
                WebhookPusher(
                    request.state.community.community_name,
                    data.api_schema
                ).match_start
            )
        )


class DemoUploadAPI(HTTPEndpoint):
    @requires("master")
    async def put(self, request: Request) -> response:
        """Used to upload a demo.

        Parameters
        ----------
        request : Request
        """

        match = request.state.community.match(request.path_params["match_id"])

        demo = Demo(match, request)
        if not demo.upload:
            return response()

        try:
            demo_status = await match.demo_status()
        except InvalidMatchID:
            raise
        else:
            if demo_status != 0:
                raise DemoAlreadyUploaded()

            await match.set_demo_status(1)

            if await demo.upload():
                background_task = None

                await match.set_demo_status(2)
            else:
                background_task = BackgroundTask(
                    request.state.community.email,
                    title="SQLMatches.com, upload failed.",
                    content=("""A upload for a demo failed due to it
                    being too large. Visit the owner panel to
                    increase this!"""),
                    link_href=(
                        Config.frontend_url + "c/{}/owner#tab2".format(
                            match.community_name
                        )
                    ),
                    link_text="{}'s owner panel.".format(
                        match.community_name
                    )
                )

                await match.set_demo_status(3)

            scoreboard = await match.scoreboard()
            data = scoreboard.api_schema

            await (CommunityCache(
                match.community_name
            ).scoreboard(request.path_params["match_id"])).set(
                data
            )

            await Sessions.websocket.emit(
                "match_update",
                data,
                room="ws_room"
            )

            await Sessions.websocket.emit(
                request.path_params["match_id"],
                data,
                room="ws_room"
            )

            return response(background=background_task)



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


from starlette.endpoints import HTTPEndpoint
from starlette.authentication import requires
from starlette.requests import Request
from starlette.background import BackgroundTask

from webargs import fields
from webargs_starlette import use_args

from ...misc import bulk_scoreboard_expire

from ...community import create_community, get_community_from_owner
from ...exceptions import InvalidCommunity, NoOwnership

from ...responses import response

from ...resources import Config, Sessions

from ...caches import CommunityCache, CommunitiesCache


class PublicCommunityAPI(HTTPEndpoint):
    @requires("community")
    async def get(self, request: Request) -> response:
        """Used to get public details on community.

        Parameters
        ----------
        request : Request

        Returns
        -------
        response
        """

        return response(
            (
                await request.state.community.public()
            ).api_schema
        )


class CommunityExistsAPI(HTTPEndpoint):
    @requires("community")
    async def get(self, request: Request) -> response:
        """Used to check if community already exists.

        Parameters
        ----------
        request : Request

        Returns
        -------
        response
        """

        return response({
            "taken": await request.state.community.exists()
        })


class CommunitySessionAPI(HTTPEndpoint):
    @requires("is_owner")
    async def get(self, request: Request) -> response:
        return response({
            "url": await request.state.community.billing_session()
        })

    @requires("is_owner")
    async def post(self, request: Request) -> response:
        return response({
            "session_id": await request.state.community.checkout_session()
        })


class CommunityOwnerAPI(HTTPEndpoint):
    @requires("is_owner")
    async def get(self, request: Request) -> response:
        """Gets community details including secrets.

        Parameters
        ----------
        request : Request
        """

        data = {}

        cache = CommunityCache(request.state.community.community_name)
        cache_get = await cache.get()
        if cache_get:
            data["community"] = cache_get
        else:
            try:
                community = await request.state.community.get()
            except InvalidCommunity:
                raise
            else:
                data["community"] = community.api_schema

                await cache.set(data["community"])

        stats_cache = cache.stats()
        stats_cache_get = await stats_cache.get()
        if stats_cache_get:
            data["stats"] = stats_cache_get
        else:
            data["stats"] = (
                await request.state.community.stats()
            ).api_schema

            await stats_cache.set(data["stats"])

        return response(data)

    @requires("is_owner")
    async def delete(self, request: Request) -> response:
        """Used to disable a community.

        Parameters
        ----------
        request : Request
        """

        await request.state.community.disable()

        await (CommunityCache(
            request.state.community.community_name
        )).expire()

        communities_cache = CommunitiesCache()
        await communities_cache.expire()
        await (communities_cache.matches()).expire()

        return response()

    @requires("is_owner")
    async def post(self, request: Request) -> response:
        """Used to regenerate a key for a community.

        Parameters
        ----------
        request : Request
        """

        return response({
            "master_api_key": await request.state.community.regenerate_master()
        })


class CommunityUpdateAPI(HTTPEndpoint):
    @use_args({"demos": fields.Bool(), "community_type": fields.Str(),
               "match_start_webhook": fields.Str(min=5, max=255),
               "round_end_webhook": fields.Str(min=5, max=255),
               "match_end_webhook": fields.Str(min=5, max=255),
               "allow_api_access": fields.Bool(),
               "email": fields.Str(max=255)})
    @requires("is_owner")
    async def post(self, request: Request, parameters: dict) -> response:
        await request.state.community.update(**parameters)

        await (CommunityCache(request.state.community.community_name)).expire()

        return response()


class CommunityOwnerMatchesAPI(HTTPEndpoint):
    @use_args({"matches": fields.List(fields.String(), required=True)})
    @requires("is_owner")
    async def delete(self, request: Request, parameters: dict) -> response:
        """Used to bulk delete matches.

        Parameters
        ----------
        request : Request
        parameters : dict

        Returns
        -------
        response
        """

        await request.state.community.delete_matches(**parameters)

        cache = CommunityCache(request.state.community.community_name)
        await (cache.matches()).expire()
        await cache.expire()

        await (CommunitiesCache().matches()).expire()

        return response(background=BackgroundTask(
            bulk_scoreboard_expire,
            community_name=request.state.community.community_name,
            matches=parameters["matches"]
        ))


class CommunityCreateAPI(HTTPEndpoint):
    @use_args({"community_name": fields.Str(required=True, max=32, min=4),
               "email": fields.Str(required=True, max=255),
               "community_type": fields.Str(),
               "demos": fields.Bool(),
               "allow_api_access": fields.Bool()})
    @requires("steam_login")
    async def post(self, request: Request, parameters: dict) -> response:
        """Used to create a community.

        Parameters
        ----------
        request : Request
        parameters : dict
        """

        model, community = await create_community(
            steam_id=request.session["steam_id"],
            **parameters
        )

        await Sessions.websocket.emit(
            "community_updates",
            model.api_schema,
            room="ws_room"
        )

        await CommunityCache(parameters["community_name"]).set(
            model.api_schema
        )
        await CommunitiesCache().expire()

        return response(model.api_schema, background=BackgroundTask(
            community.email,
            title="SQLMatches.com welcomes you, {}!".format(
                model.community_name
            ),
            content=("""Thanks for creating a community.
            Need to upload demos larger then {} MB? Consider buying a larger
            max upload on the owner panel.""".format(Config.free_upload_size)),
            link_href=Config.frontend_url + "c/{}/owner#tab2".format(
                model.community_name
            ),
            link_text="{}'s owner panel.".format(model.community_name)
        ))

    @requires("steam_login")
    async def get(self, request: Request) -> response:
        """Used to valid if user owns a community.

        Parameters
        ----------
        request : Request

        Returns
        -------
        response
        """

        try:
            community, banned, _ = await get_community_from_owner(
                request.session["steam_id"]
            )

            community_name = community.community_name
        except NoOwnership:
            community_name = None
            banned = False

        return response({
            "community_name": community_name,
            "banned": banned,
            "steam_id": request.session["steam_id"]
        })



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import aiofiles

from os import path
from zipfile import ZipFile

from starlette.endpoints import HTTPEndpoint
from starlette.authentication import requires
from starlette.requests import Request
from starlette.background import BackgroundTask

from webargs import fields
from webargs_starlette import use_args

from ...responses import response, error_response
from ...resources import Config, Sessions
from ...communities import ban_communities
from ...caches import CommunitiesCache, VersionCache, VersionsCache
from ...misc import bulk_community_expire
from ...version import Version


class CommunitiesAdminAPI(HTTPEndpoint):
    @use_args({"communities": fields.List(fields.String(), required=True)})
    @requires("root_login")
    async def delete(self, request: Request, parameters: dict) -> response:
        """Used to ban communities.

        Parameters
        ----------
        request : Request
        parameters : dict

        Returns
        -------
        response
        """

        await ban_communities(**parameters)

        await CommunitiesCache().expire()

        return response(background=BackgroundTask(
            bulk_community_expire,
            **parameters
        ))


class SavePluginAPI(HTTPEndpoint):
    @use_args({"zip_url": fields.Url(required=True)})
    @requires("root_login")
    async def post(self, request: Request, paramters: dict) -> response:
        """Used to save a version of the plugins locally.

        Parameters
        ----------
        request : Request
        paramters : dict

        Returns
        -------
        response
        """

        # Temp
        return error_response("Not implemented")

        async with Sessions.aiohttp.get(paramters["zip_url"]) as resp:
            if resp.status != 200:
                return error_response(
                    "Unable to download file",
                    status_code=resp.status
                )

            zip_pathway = path.join(Config.plugin_dir, "plugins.zip")

            async with aiofiles.open(zip_pathway, "w+") as file_:
                await file_.truncate()
                await file_.write(await resp.read())

            with ZipFile(zip_pathway) as zf:
                zf.extractall(path.join(Config.plugin_dir, "extracted"))

            return response("Files unzipped and cached")


class AdminAPI(HTTPEndpoint):
    @use_args({"major": fields.Int(required=True),
               "minor": fields.Int(required=True),
               "patch": fields.Int(required=True),
               "message": fields.String(min=6, max=64, required=True)})
    @requires("root_login")
    async def post(self, request: Request, parameters: dict) -> response:
        """Used to create or update a message.

        Parameters
        ----------
        request : Request
        parameters : dict

        Returns
        -------
        response
        """

        await Version(
            parameters["major"],
            parameters["minor"],
            parameters["patch"]
        ).save(parameters["message"])

        await VersionCache(
            parameters["major"],
            parameters["minor"],
            parameters["patch"]
        ).set(parameters["message"])

        await VersionsCache().expire()

        return response()



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


import asyncio
import logging
import aiofiles

from starlette.requests import Request
from sqlalchemy.sql import and_, select, func
from typing import Any
from datetime import datetime

from backblaze.exceptions import BackblazeException

from os import path
from backblaze.settings import PartSettings, UploadSettings

from .community.match import Match
from .resources import Config, Sessions
from .settings import B2UploadSettings, LocalUploadSettings
from .tables import scoreboard_total_table, community_table


class Demo:
    def __init__(self, match: Match, request: Request = None) -> None:
        """Wrapper for demo uploading.

        Parameters
        ----------
        match : Match
        request : Request
        """

        self.match = match
        self.request = request

        if Config.upload_type == B2UploadSettings:
            self.upload = self.__b2_upload
            self.delete = self.__b2_delete
        elif Config.upload_type == LocalUploadSettings:
            self.upload = self.__local_upload
            self.delete = self.__local_delete
        else:
            self.upload = None
            self.delete = None

    @property
    def __file_name(self) -> str:
        return self.match.match_id + Config.demo_extension

    @property
    def __demo_pathway(self) -> str:
        pathway = path.join(
            Config.demo_pathway,
            self.__file_name
        )

        if Config.upload_type == B2UploadSettings:
            # Backblaze only works with '/'
            return pathway.replace("\\", "/")

        return pathway

    @property
    def __where_statement(self) -> Any:
        return and_(
            scoreboard_total_table.c.match_id == self.match.match_id,
            scoreboard_total_table.c.community_name ==
            self.match.community_name
        )

    async def __invalid_upload(self, total_size: int) -> bool:
        """Used to see if upload size is invalid.

        Parameters
        ----------
        total_size : int
            Total size in bytes.

        Returns
        -------
        bool
            If invalid or not.
        """

        if total_size == 0:
            return False

        allow_max_upload = bool(await Sessions.database.fetch_val(
            select([func.count()]).select_from(community_table).where(
                and_(
                    community_table.c.community_name ==
                    self.match.community_name,
                    community_table.c.subscription_expires >= datetime.now()
                )
            )
        ))

        size_in_mb = total_size / 1000000

        return (
            size_in_mb > Config.max_upload_size if allow_max_upload
            else size_in_mb > Config.free_upload_size
        )

    async def __update_value(self, **kwargs) -> None:
        await Sessions.database.execute(
            scoreboard_total_table.update().values(
                **kwargs
            ).where(
                self.__where_statement
            )
        )

    async def __local_delete(self) -> bool:
        """Used to delete a local demo.

        Returns
        -------
        bool
            If demo deleted.
        """

        try:
            async with aiofiles.open(self.__demo_pathway, "wb+") as f:
                await f.truncate()
        except IOError:
            return False
        else:
            await self.__update_value(demo_status=4)
            return True

    async def __b2_delete(self) -> bool:
        b2_id = await Sessions.database.fetch_val(
            select([scoreboard_total_table.c.b2_id]).select_from(
                scoreboard_total_table
            ).where(self.__where_statement)
        )

        if not b2_id:
            return False

        try:
            await (Sessions.bucket.file(b2_id)).delete(self.__demo_pathway)
        except BackblazeException as error:
            logging.warning("Backblaze failed because of\n{}".format(error))
            return False
        else:
            await self.__update_value(demo_status=4)
            return True

    async def __local_upload(self) -> bool:
        """Used to upload file locally.

        Returns
        -------
        bool
            If demo uploaded.
        """

        assert self.request

        async with aiofiles.open(self.__demo_pathway, "wb+") as f:
            total_size = 0
            async for chunk in self.request.stream():
                await f.write(chunk)
                total_size += len(chunk)

                await asyncio.sleep(Config.upload_delay)

            if await self.__invalid_upload(total_size):
                await f.truncate()
                return False
            else:
                return True

    async def __b2_upload(self) -> bool:
        """Used to upload demo to b2.

        Returns
        -------
        bool
            If demo uploaded.
        """

        assert self.request

        content_type = "application/octet-stream"

        model, file = await Sessions.bucket.create_part(PartSettings(
            self.__demo_pathway,
            content_type=content_type
        ))

        parts = file.parts()

        chunked = b""
        total_size = 0
        async for chunk in self.request.stream():
            chunked += chunk

            if len(chunked) >= 5000000:
                await parts.data(chunked)

                total_size += len(chunked)
                chunked = b""

            await asyncio.sleep(Config.upload_delay)

        if chunked:
            if parts.part_number == 0:
                await file.cancel()

                model, _ = await Sessions.bucket.upload(UploadSettings(
                    self.__demo_pathway,
                    content_type=content_type
                ), chunked)

                await self.__update_value(b2_id=model.file_id)

                return True
            else:
                await parts.data(chunked)
                total_size += len(chunked)

        if await self.__invalid_upload(total_size):
            await file.cancel()
            return False
        else:
            await parts.finish()

            await self.__update_value(b2_id=model.file_id)

            return True



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


from typing import Any, Dict, List

from datetime import datetime

from sqlalchemy.sql import select, and_, func

from asyncio import sleep

from ..on_conflict import (
    on_scoreboard_conflict,
    on_user_conflict,
    on_statistic_conflict
)
from ..tables import scoreboard_total_table, scoreboard_table, user_table
from ..resources import Sessions

from .models import ScoreboardModel
from ..exceptions import InvalidMatchID


class Match:
    def __init__(self, match_id: str, community_name: str) -> None:
        """Handles interactions with a match

        Paramters
        ---------
        match_id: str
            ID of match
        community_name: str
            Name of community.
        """

        self.match_id = match_id
        self.community_name = community_name

    async def set_demo_status(self, status) -> None:
        """Sets demo status to given value.

        Raises
        ------
        InvalidMatchID
            Raised when match ID is invalid.
        """

        try:
            query = scoreboard_total_table.update().values(
                demo_status=status
            ).where(
                and_(
                    scoreboard_total_table.c.match_id == self.match_id,
                    scoreboard_total_table.c.community_name
                    == self.community_name
                )
            )

            await Sessions.database.execute(query=query)
        except Exception:
            raise InvalidMatchID()

    async def demo_status(self) -> int:
        """
        Gets demo status.

        Raises
        ------
        InvalidMatchID
            Raised when match ID is invalid.

        Returns
        -------
        int
            Status of demo.
        """

        query = select([scoreboard_total_table.c.demo_status]).select_from(
            scoreboard_total_table.join(
                scoreboard_table,
                scoreboard_table.c.match_id ==
                scoreboard_total_table.c.match_id
            )
        ).where(
            and_(
                scoreboard_total_table.c.match_id == self.match_id,
                scoreboard_total_table.c.community_name == self.community_name
            )
        ).distinct()

        demo_status = await Sessions.database.fetch_val(query=query)
        if demo_status is not None:
            return demo_status
        else:
            raise InvalidMatchID()

    async def exists(self) -> bool:
        """Checks if the match exists.

        Returns
        -------
        bool
        """

        query = select([func.count()]).where(
            and_(
                scoreboard_total_table.c.match_id == self.match_id,
                scoreboard_total_table.c.community_name == self.community_name
            )
        )

        return await Sessions.database.fetch_val(query=query) > 0

    async def update(self, team_1_score: int, team_2_score: int,
                     players: List[Dict[str, Dict[str, Any]]] = None,
                     team_1_side: int = None, team_2_side: int = None,
                     end: bool = False) -> None:
        """Updates match details.

        Raises
        ------
        InvalidMatchID
            Raised when match ID is invalid.
        """

        if not await self.exists():
            raise InvalidMatchID()

        team_sides = {}

        if team_1_side is not None:
            team_sides["team_1_side"] = team_1_side

        if team_2_side is not None:
            team_sides["team_2_side"] = team_2_side

        query = scoreboard_total_table.update().values(
            team_1_score=team_1_score,
            team_2_score=team_2_score,
            status=0 if end else 1,
            **team_sides
        ).where(
            and_(
                scoreboard_total_table.c.match_id == self.match_id,
                scoreboard_total_table.c.community_name == self.community_name
            )
        )

        await Sessions.database.execute(query=query)

        if players:
            scoreboard = []
            scoreboard_append = scoreboard.append

            users = []
            users_append = users.append

            statistics = []
            statistics_append = statistics.append

            now = datetime.now()

            for player in players:
                scoreboard_append({
                    "match_id": self.match_id,
                    "steam_id": player["steam_id"],
                    "team": player["team"],
                    "alive": player["alive"],
                    "ping": player["ping"],
                    "kills": player["kills"],
                    "headshots": player["headshots"],
                    "assists": player["assists"],
                    "deaths": player["deaths"],
                    "shots_fired": player["shots_fired"],
                    "shots_hit": player["shots_hit"],
                    "mvps": player["mvps"],
                    "score": player["score"],
                    "disconnected": player["disconnected"]
                })

                users_append({
                    "steam_id": player["steam_id"],
                    "name": player["name"],
                    "timestamp": now
                })

                statistics_append({
                    "community_name": self.community_name,
                    "steam_id": player["steam_id"],
                    "kills": player["kills"],
                    "headshots": player["headshots"],
                    "assists": player["assists"],
                    "deaths": player["deaths"],
                    "shots_fired": player["shots_fired"],
                    "shots_hit": player["shots_hit"],
                    "mvps": player["mvps"]
                })

                await sleep(0.000001)

            await Sessions.database.execute_many(
                query=on_user_conflict(),
                values=users
            )

            await Sessions.database.execute_many(
                query=on_scoreboard_conflict(),
                values=scoreboard
            )

            await Sessions.database.execute_many(
                query=on_statistic_conflict(),
                values=statistics
            )

    async def end(self) -> None:
        """Sets match status to 0

        Raises
        ------
        InvalidMatchID
            Raised when match ID is invalid.
        """

        query = scoreboard_total_table.update().values(
            status=0
        ).where(
            and_(
                scoreboard_total_table.c.match_id == self.match_id,
                scoreboard_total_table.c.community_name == self.community_name
            )
        )

        try:
            await Sessions.database.execute(query=query)
        except Exception:
            raise InvalidMatchID()

    async def scoreboard(self) -> ScoreboardModel:
        """Gets scoreboard data.

        Returns
        ------
        ScoreboardModel
            Holds scoreboard data.
        """

        query = select([
            scoreboard_total_table.c.timestamp,
            scoreboard_total_table.c.status,
            scoreboard_total_table.c.map,
            scoreboard_total_table.c.demo_status,
            scoreboard_total_table.c.team_1_name,
            scoreboard_total_table.c.team_2_name,
            scoreboard_total_table.c.team_1_score,
            scoreboard_total_table.c.team_2_score,
            scoreboard_total_table.c.team_1_side,
            scoreboard_total_table.c.team_2_side,
            scoreboard_total_table.c.community_name,
            user_table.c.steam_id,
            user_table.c.name,
            scoreboard_table.c.team,
            scoreboard_table.c.alive,
            scoreboard_table.c.ping,
            scoreboard_table.c.kills,
            scoreboard_table.c.headshots,
            scoreboard_table.c.assists,
            scoreboard_table.c.deaths,
            scoreboard_table.c.shots_fired,
            scoreboard_table.c.shots_hit,
            scoreboard_table.c.mvps,
            scoreboard_table.c.score,
            scoreboard_table.c.disconnected
        ]).select_from(
            scoreboard_total_table.join(
                scoreboard_table,
                scoreboard_table.c.match_id ==
                scoreboard_total_table.c.match_id
            ).join(
                user_table,
                user_table.c.steam_id == scoreboard_table.c.steam_id
            )
        ).where(
            and_(
                scoreboard_total_table.c.match_id == self.match_id,
                scoreboard_total_table.c.community_name == self.community_name
            )
        ).order_by(scoreboard_table.c.score.desc())

        scoreboard_data = {
            "match": None,
            "team_1": [],
            "team_2": []
        }

        team_1_append = scoreboard_data["team_1"].append
        team_2_append = scoreboard_data["team_2"].append

        async for row in Sessions.database.iterate(query=query):
            if not scoreboard_data["match"]:
                scoreboard_data["match"] = {
                    "match_id": self.match_id,
                    "timestamp": row["timestamp"],
                    "status": row["status"],
                    "demo_status": row["demo_status"],
                    "map": row["map"],
                    "team_1_name": row["team_1_name"],
                    "team_2_name": row["team_2_name"],
                    "team_1_score": row["team_1_score"],
                    "team_2_score": row["team_2_score"],
                    "team_1_side": row["team_1_side"],
                    "team_2_side": row["team_2_side"],
                    "community_name": row["community_name"]
                }

            team_append = team_1_append if row["team"] == 0 else team_2_append

            team_append({
                "steam_id": row["steam_id"],
                "name": row["name"],
                "team": row["team"],
                "alive": row["alive"],
                "ping": row["ping"],
                "kills": row["kills"],
                "headshots": row["headshots"],
                "assists": row["assists"],
                "deaths": row["deaths"],
                "shots_fired": row["shots_fired"],
                "shots_hit": row["shots_hit"],
                "mvps": row["mvps"],
                "score": row["score"],
                "disconnected": row["disconnected"]
            })

        if scoreboard_data["match"]:
            return ScoreboardModel(**scoreboard_data)
        else:
            raise InvalidMatchID()



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from os import path, mkdir

from .exceptions import UnSupportedEngine
from .resources import Config


class DatabaseSettings:
    def __init__(
                self,
                username: str,
                password: str,
                server: str,
                port: int,
                database: str,
                engine: str = "mysql"
                ) -> None:
        """Database settings.

        Parameters
        ----------
        username : str
        password : str
        server : str
        port : int
        database : str
        engine : str, optional
            by default "mysql"

        Raises
        ------
        UnSupportedEngine
        """

        self.username = username
        self.password = password
        self.server = server
        self.port = port
        self.database = database
        self.engine = engine

        if engine == "mysql":
            self.alchemy_engine = "pymysql"
        elif engine == "sqlite":
            self.alchemy_engine = "sqlite3"
        elif engine == "postgresql":
            self.alchemy_engine = "psycopg2"
        else:
            raise UnSupportedEngine()

        Config.db_engine = engine


class __Extension:
    def __init__(self, extension: str = ".dem.bz2") -> None:
        self.extension = extension


class B2UploadSettings(__Extension):
    def __init__(self, key_id: str, application_key: str,
                 bucket_id: str, pathway: str, cdn_url: str,
                 *args, **kwargs) -> None:
        """B2 Settings

        Parameters
        ----------
        key_id: str
            B2 key ID.
        application_key: str
            B2 app key.
        bucket_id: str
            Bucket to upload demos to.
        pathway: str
            Pathway to store demos to.
        cdn_url: str
            URL to access files.
        extension: str,
            by default ".dem.bz2"
        """

        super().__init__(*args, **kwargs)

        self.key_id = key_id
        self.application_key = application_key
        self.bucket_id = bucket_id
        self.cdn_url = cdn_url if cdn_url[-1:] == "/" else cdn_url + "/"

        if pathway[-1:] == "/":
            pathway = pathway[:-1]

        if pathway[0] == "/":
            pathway = pathway[1:]

        self.pathway = pathway


class LocalUploadSettings(__Extension):
    def __init__(self, pathway: str = None, *args, **kwargs) -> None:
        """Used to upload demos locally, not recommend!

        Parameters
        ----------
        pathway : str
            Pathway to upload to from package location, by default None
        extension: str,
            by default ".dem.bz2"
        """

        super().__init__(*args, **kwargs)

        if pathway:
            self.pathway = pathway
        else:
            self.pathway = path.join(
                path.abspath(path.dirname(__file__)),
                "demos"
            )

        if not path.exists(self.pathway):
            mkdir(self.pathway)


class StripeSettings:
    def __init__(self, api_key: str, price_id: str,
                 testing: bool = False) -> None:
        """Used to set stripe settings.

        Parameters
        ----------
        api_key : str
            Don't include 'sk_test_' or 'sk_live_'
        price_id : str
            Stripe price ID, created on dashboard.
        testing : bool, optional
            by default False
        """

        self.api_key = api_key
        self.price_id = price_id
        self.testing = testing


class SmtpSettings:
    def __init__(self, hostname: str, port: int, use_tls: bool = True,
                 username: str = None, password: str = None) -> None:
        """SMTP Connection settings.

        Parameters
        ----------
        hostname : str
        port : int
        use_tls : bool, optional
        username : str, optional
        password : str, optional
        """

        self.hostname = hostname
        self.port = port
        self.use_tls = use_tls
        self.username = username
        self.password = password


class WebhookSettings:
    def __init__(self, timeout: float = 3.0, match_start: str = None,
                 match_end: str = None, round_end: str = None,
                 key: str = None) -> None:
        """Used to configure webhook.

        Parameters
        ----------
        timeout : float, optional
            by default 3.0
        match_start : str, optional
            Global webhook for match start, by default None
        match_end : str, optional
            Global webhook for match end, by default None
        round_end : str, optional
            Global webhook for round end, by default None
        key : str, optional
            Global webhook key.
        """

        self.timeout = timeout
        self.match_start = match_start
        self.match_end = match_end
        self.round_end = round_end
        self.key = key



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


from typing import Tuple

from .card import Card

from .models import SubscriptionModel, CardModel, SessionModel
from .subscription import Subscription


class Customer:
    def __init__(self, id: str, context: object) -> None:
        self.id = id
        self._context = context

    async def create_subscription(self, unit_amount_decimal: float,
                                  currency: str,
                                  product_id: str,
                                  interval_days: int,
                                  cancel_at_period_end: bool = False,
                                  ) -> Tuple[SubscriptionModel, Subscription]:
        """Used to create a subscription.

        Parameters
        ----------
        unit_amount_decimal : float
        currency : str
        product_id : str
        interval_days : int
        cancel_at_period_end : bool, optional
            by default False

        Returns
        -------
        SubscriptionModel
        Subscription
        """

        data = await self._context._post(
            "subscriptions",
            data={
                "customer": self.id,
                "cancel_at_period_end": cancel_at_period_end,
                "items[0][price_data][unit_amount_decimal]":
                unit_amount_decimal,
                "items[0][price_data][currency]": currency,
                "items[0][price_data][product]": product_id,
                "items[0][price_data][recurring][interval]": "day",
                "items[0][price_data][recurring][interval_count]":
                interval_days
            }
        )

        return (
            SubscriptionModel(**data),
            self._context.subscription(data["id"])
        )

    def card(self, id: str) -> Card:
        """Used to interact with a card.

        Parameters
        ----------
        id : str

        Returns
        -------
        Card
        """

        return Card(self.id, id, self._context)

    async def _token(self, number: str, exp_month: int,
                     exp_year: int, cvc: int,
                     name: str) -> str:
        """Used to create token.

        Parameters
        ----------
        number : str
        exp_month : int
        exp_year : int
        cvc : int
        name : str

        Returns
        -------
        str
        """

        return (await self._context._post(
            "tokens",
            data={
                "[card]number": number,
                "[card]exp_month": exp_month,
                "[card]exp_year": exp_year,
                "[card]cvc": cvc,
                "[card]name": name
            }
        ))["id"]

    async def create_card(self, number: str, exp_month: int,
                          exp_year: int, cvc: int,
                          name: str
                          ) -> Tuple[CardModel, Card]:
        """Used to create a card.

        Parameters
        ----------
        number : str
        exp_month : int
        exp_year : int
            Expects full year format.
        cvc : int
        name : str

        Returns
        -------
        CardModel
        Card
        """

        data = await self._context._post(
            "customers/{}/sources".format(self.id),
            data={
                "source": await self._token(
                    number, exp_month, exp_year,
                    cvc, name
                )
            }
        )

        return CardModel(**data), self.card(data["id"])

    async def create_billing_session(self, return_url: str) -> SessionModel:
        """Used to create a billing session.

        Parameters
        ----------
        return_url : str

        Returns
        -------
        SessionModel
        """

        data = await self._context._post(
            "billing_portal/sessions",
            data={
                "customer": self.id,
                "return_url": return_url
            }
        )

        return SessionModel(**data)

    async def create_checkout_session(self, cancel_url: str, success_url: str,
                                      price_id: str,
                                      payment_method_types: str = "card",
                                      mode: str = "subscription",
                                      quantity: int = 1) -> str:
        """Used to create a checkout session.

        Parameters
        ----------
        cancel_url : str
        success_url : str
        price_id : str
        payment_method_types : str, optional
            by default "card"
        mode : str, optional
            by default "subscription"
        quantity : int, optional
            by default 1

        Returns
        -------
        str
            Checkout session ID.
        """

        data = await self._context._post(
            "checkout/sessions",
            data={
                "cancel_url": cancel_url,
                "success_url": success_url,
                "line_items[0][price]": price_id,
                "line_items[0][quantity]": quantity,
                "payment_method_types[0]": payment_method_types,
                "mode": mode,
                "customer": self.id
            }
        )

        return data["id"]



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request
from starlette.background import BackgroundTask
from starlette.authentication import requires

from ..resources import Config
from ..responses import error_response, response
from ..community import stripe_customer_to_community
from ..exceptions import InvalidCustomer
from ..caches import CommunityCache


class PaymentFailedWebhook(HTTPEndpoint):
    @requires("stripe_webhook")
    async def post(self, request: Request) -> response:
        json = await request.json()

        if json["type"] != "charge.failed":
            return error_response("charge.failed expected")

        try:
            community = await stripe_customer_to_community(
                json["data"]["object"]["customer"]
            )
        except InvalidCustomer:
            raise
        else:
            await CommunityCache(community.community_name).expire()

            return response(background=BackgroundTask(
                community.email,
                title="SQLMatches - Payment Failed!",
                content=("""Your payment was declined! To fix this
                please go to your owner panel."""),
                link_href="{}'s owner panel".format(community.community_name),
                link_text="{}c/{}/owner#tab2".format(
                    Config.frontend_url, community.community_name
                )
            ))


class PaymentSuccessWebhook(HTTPEndpoint):
    @requires("stripe_webhook")
    async def post(self, request: Request) -> response:
        json = await request.json()

        if json["type"] != "charge.succeeded":
            return error_response("charge.succeeded expected")

        try:
            community = await stripe_customer_to_community(
                json["data"]["object"]["customer"]
            )
        except InvalidCustomer:
            raise
        else:
            await CommunityCache(community.community_name).expire()

            await community.update_subscription_expire()

            return response(background=BackgroundTask(
                community.email,
                title="SQLMatches - Invoice",
                content=("""Thanks for supporting SQLMatches,
                your receipt can be found below."""),
                link_href=json["data"]["object"]["receipt_url"],
                link_text="{}'s Receipt".format(community.community_name)
            ))



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


from typing import Tuple
from aiohttp import ClientResponse
from functools import wraps

from ..resources import Sessions

from .models import CustomerModel, ProductModel
from .subscription import Subscription
from .customer import Customer


def add_headers(func):
    @wraps(func)
    def _add(*args, **kwargs):
        auth = (args[0]._context.authorization if hasattr(args[0], "_context")
                else args[0].authorization)

        if "headers" in kwargs:
            kwargs["headers"]["Authorization"] = auth
            kwargs["headers"]["Content-Type"] = \
                "application/x-www-form-urlencoded"
        else:
            kwargs["headers"] = {
                "Authorization": auth,
                "Content-Type": "application/x-www-form-urlencoded"
            }

        return func(*args, **kwargs)

    return _add


class Stripe:
    BASE_URL = "https://api.stripe.com/v1/"

    def __init__(self, api_key: str, testing: bool = False) -> None:
        """Used to interact with Stripe.

        Parameters
        ----------
        api_key : str
        testing : bool, optional
            Enable test mode, by default False
        """

        self.authorization = "Bearer sk_{}_{}".format(
            "test" if testing else "live", api_key
        )

    async def __handle(self, resp: ClientResponse) -> dict:
        resp.raise_for_status()
        return await resp.json()

    @add_headers
    async def _post(self, path: str, *args, **kwargs) -> dict:
        """Used to post to Stripe.

        Parameters
        ----------
        path : str
            Path to append to stripe API route.

        Returns
        -------
        dict
        """

        async with Sessions.aiohttp.post(url=self.BASE_URL + path,
                                         *args, **kwargs) as resp:
            return await self.__handle(resp)

    @add_headers
    async def _delete(self, path: str, *args, **kwargs) -> dict:
        """Used to delete something from Stripe.

        Parameters
        ----------
        path : str
            Path to append to stripe API route.

        Returns
        -------
        dict
        """

        async with Sessions.aiohttp.delete(url=self.BASE_URL + path,
                                           *args, **kwargs) as resp:
            return await self.__handle(resp)

    @add_headers
    async def _get(self, path: str, *args, **kwargs) -> dict:
        """Used to get Stripe data.

        Parameters
        ----------
        path : str
            Path to append to stripe API route.

        Returns
        -------
        dict
        """

        async with Sessions.aiohttp.get(url=self.BASE_URL + path,
                                        *args, **kwargs) as resp:
            return await self.__handle(resp)

    def subscription(self, id: str) -> Subscription:
        """Interact with a subscription.

        Parameters
        ----------
        id : str

        Returns
        -------
        Subscription
        """

        return Subscription(id, self)

    def customer(self, id: str) -> Customer:
        """Used to interact with customer.

        Parameters
        ----------
        id : str

        Returns
        -------
        Customer
        """

        return Customer(id, self)

    async def create_product(self, name: str) -> ProductModel:
        """Used to create a product.

        Parameters
        ----------
        name : str

        Returns
        -------
        ProductModel
        """

        data = await self._post("products", data={
            "name": name
        })

        return ProductModel(**data)

    async def create_customer(self, **kwargs
                              ) -> Tuple[CustomerModel, Customer]:
        """Used to create a customer.

        Returns
        -------
        CustomerModel
        """

        data = await self._post("customers", data=kwargs)

        return CustomerModel(**data), self.customer(data["id"])



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2021 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


from typing import Dict, List, Any


class RecurringModel:
    aggregate_usage: None
    interval: str
    interval_count: int
    usage_type: str
    metadata: Dict[str, Any]

    def __init__(self, **kwargs) -> None:
        self.aggregate_usage = kwargs.get("aggregate_usage")
        self.interval = kwargs.get("interval")
        self.interval_count = kwargs.get("interval_count")
        self.usage_type = kwargs.get("usage_type")
        self.metadata = kwargs.get("metadata")


class PriceModel:
    id: str
    object: str
    active: bool
    billing_scheme: str
    created: int
    currency: str
    livemode: bool
    lookup_key: None
    nickname: None
    product: str
    recurring: RecurringModel
    tiers_mode: None
    transform_quantity: None
    type: str
    unit_amount: int
    unit_amount_decimal: int
    metadata: Dict[str, Any]

    def __init__(self, **kwargs) -> None:
        self.id = kwargs.get("id")
        self.object = kwargs.get("object")
        self.active = kwargs.get("active")
        self.billing_scheme = kwargs.get("billing_scheme")
        self.created = kwargs.get("created")
        self.currency = kwargs.get("currency")
        self.livemode = kwargs.get("livemode")
        self.lookup_key = kwargs.get("lookup_key")
        self.nickname = kwargs.get("nickname")
        self.product = kwargs.get("product")
        self.recurring = RecurringModel(
            **kwargs["recurring"]
        ) if "recurring" in kwargs else None
        self.tiers_mode = kwargs.get("tiers_mode")
        self.transform_quantity = kwargs.get("transform_quantity")
        self.type = kwargs.get("type")
        self.unit_amount = kwargs.get("unit_amount")
        self.unit_amount_decimal = kwargs.get("unit_amount_decimal")
        self.metadata = kwargs.get("metadata")


class DatumModel:
    id: str
    object: str
    billing_thresholds: None
    created: int
    price: PriceModel
    quantity: int
    subscription: str
    tax_rates: List[Any]
    metadata: Dict[str, Any]

    def __init__(self, **kwargs) -> None:
        self.id = kwargs.get("id")
        self.object = kwargs.get("object")
        self.billing_thresholds = kwargs.get("billing_thresholds")
        self.created = kwargs.get("created")
        self.price = PriceModel(
            **kwargs["price"]
        ) if "price" in kwargs else None
        self.quantity = kwargs.get("quantity")
        self.subscription = kwargs.get("subscription")
        self.tax_rates = kwargs.get("tax_rates")
        self.metadata = kwargs.get("metadata")


class ItemsModel:
    object: str
    data: List[DatumModel]
    has_more: bool
    url: str
    metadata: Dict[str, Any]

    def __init__(self, **kwargs) -> None:
        self.object = kwargs.get("kwargs")
        self.data = [
            DatumModel(**datum) for datum in kwargs["data"]
        ] if "data" in kwargs else None
        self.has_more = kwargs.get("has_more")
        self.url = kwargs.get("url")
        self.metadata = kwargs.get("metadata")


class InvoiceSettingsModel:
    custom_fields: str
    default_payment_method: None
    footer: str
    metadata: Dict[str, Any]

    def __init__(self, **kwargs) -> None:
        self.custom_fields = kwargs.get("custom_fields")
        self.default_payment_method = kwargs.get("default_payment_method")
        self.footer = kwargs.get("footer")
        self.metadata = kwargs.get("metadata")


class CustomerModel:
    id: str
    object: str
    address: str
    balance: int
    created: int
    currency: str
    default_source: str
    delinquent: bool
    description: str
    discount: float
    email: str
    invoice_prefix: str
    invoice_settings: InvoiceSettingsModel
    livemode: bool
    name: str
    next_invoice_sequence: int
    phone: str
    preferred_locales: List[Any]
    shipping: str
    tax_exempt: str
    metadata: Dict[str, Any]

    def __init__(self, **kwargs) -> None:
        self.id = kwargs.get("id")
        self.object = kwargs.get("object")
        self.address = kwargs.get("address")
        self.balance = kwargs.get("balance")
        self.created = kwargs.get("created")
        self.currency = kwargs.get("currency")
        self.default_source = kwargs.get("default_source")
        self.delinquent = kwargs.get("delinquent")
        self.description = kwargs.get("description")
        self.discount = kwargs.get("discount")
        self.email = kwargs.get("email")
        self.invoice_prefix = kwargs.get("invoice_prefix")
        self.invoice_settings = kwargs.get("invoice_settings")
        self.livemode = kwargs.get("livemode")
        self.name = kwargs.get("name")
        self.next_invoice_sequence = kwargs.get("next_invoice_sequence")
        self.phone = kwargs.get("phone")
        self.preferred_locales = kwargs.get("preferred_locales")
        self.shipping = kwargs.get("shipping")
        self.tax_exempt = kwargs.get("tax_exempt")
        self.metadata = kwargs.get("metadata")


class CardModel:
    id: str
    object: str
    address_city: str
    address_country: str
    address_line1: str
    address_line1_check: str
    address_line2: str
    address_state: str
    address_zip: int
    address_zip_check: str
    brand: str
    country: str
    customer: str
    cvc_check: str
    dynamic_last4: str
    exp_month: int
    exp_year: int
    fingerprint: str
    funding: str
    last4: int
    metadata: Dict[str, Any]
    name: str
    tokenization_method: str

    def __init__(self, **kwargs) -> None:
        self.id = kwargs.get("id")
        self.object = kwargs.get("object")
        self.address_city = kwargs.get("address_city")
        self.address_country = kwargs.get("address_country")
        self.address_line1 = kwargs.get("address_line1")
        self.address_line1_check = kwargs.get("address_line1_check")
        self.address_line2 = kwargs.get("address_line2")
        self.address_state = kwargs.get("address_state")
        self.address_zip = kwargs.get("address_zip")
        self.address_zip_check = kwargs.get("address_zip_check")
        self.brand = kwargs.get("brand")
        self.country = kwargs.get("country")
        self.customer = kwargs.get("customer")
        self.cvc_check = kwargs.get("cvc_check")
        self.dynamic_last4 = kwargs.get("dynamic_last4")
        self.exp_month = kwargs.get("exp_month")
        self.exp_year = kwargs.get("exp_year")
        self.fingerprint = kwargs.get("fingerprint")
        self.funding = kwargs.get("funding")
        self.last4 = kwargs.get("last4")
        self.metadata = kwargs.get("metadata")
        self.name = kwargs.get("name")
        self.tokenization_method = kwargs.get("tokenization_method")


class ProductModel:
    id: str
    object: str
    active: bool
    attributes: List[str]
    created: int
    description: str
    images: List[Any]
    livemode: bool
    metadata: Dict[str, Any]
    name: str
    statement_descriptor: None
    type: str
    unit_label: None
    updated: int

    def __init__(self, **kwargs) -> None:
        self.id = kwargs.get("id")
        self.object = kwargs.get("object")
        self.active = kwargs.get("active")
        self.type = kwargs.get("type")
        self.attributes = kwargs.get("attributes")
        self.created = kwargs.get("created")
        self.description = kwargs.get("description")
        self.images = kwargs.get("images")
        self.livemode = kwargs.get("livemode")
        self.metadata = kwargs.get("metadata")
        self.name = kwargs.get("name")
        self.statement_descriptor = kwargs.get("statement_descriptor")
        self.unit_label = kwargs.get("unit_label")
        self.updated = kwargs.get("updated")


class PlanModel:
    id: str
    object: str
    active: bool
    aggregate_usage: None
    amount: int
    amount_decimal: int
    billing_scheme: str
    created: int
    currency: str
    interval: str
    interval_count: int
    livemode: bool
    metadata: Dict[str, Any]
    nickname: str
    product: str
    tiers_mode: str
    transform_usage: None
    trial_period_days: int
    usage_type: str

    def __init__(self, **kwargs) -> None:
        self.id = kwargs.get("id")
        self.object = kwargs.get("object")
        self.active = kwargs.get("active")
        self.aggregate_usage = kwargs.get("aggregate_usage")
        self.amount = kwargs.get("amount")
        self.amount_decimal = kwargs.get("amount_decimal")
        self.billing_scheme = kwargs.get("billing_scheme")
        self.created = kwargs.get("created")
        self.currency = kwargs.get("currency")
        self.interval = kwargs.get("interval")
        self.interval_count = kwargs.get("interval_count")
        self.livemode = kwargs.get("livemode")
        self.metadata = kwargs.get("metadata")
        self.nickname = kwargs.get("nickname")
        self.product = kwargs.get("product")
        self.tiers_mode = kwargs.get("tiers_mode")
        self.transform_usage = kwargs.get("transform_usage")
        self.trial_period_days = kwargs.get("trial_period_days")
        self.usage_type = kwargs.get("usage_type")


class RecurringModel:
    aggregate_usage: None
    interval: str
    interval_count: int
    trial_period_days: int
    usage_type: str

    def __init__(self, **kwargs) -> None:
        self.aggregate_usage = kwargs.get("aggregate_usage")
        self.interval = kwargs.get("interval")
        self.interval_count = kwargs.get("interval_count")
        self.trial_period_days = kwargs.get("trial_period_days")
        self.usage_type = kwargs.get("usage_type")


class PriceModel:
    id: str
    object: str
    active: bool
    billing_scheme: str
    created: int
    currency: str
    livemode: bool
    lookup_key: str
    metadata: Dict[str, Any]
    nickname: str
    product: str
    recurring: RecurringModel
    tiers_mode: bool
    transform_quantity: None
    type: str
    unit_amount: int
    unit_amount_decimal: int

    def __init__(self, **kwargs) -> None:
        self.id = kwargs.get("id")
        self.object = kwargs.get("object")
        self.active = kwargs.get("active")
        self.billing_scheme = kwargs.get("billing_scheme")
        self.created = kwargs.get("created")
        self.currency = kwargs.get("currency")
        self.livemode = kwargs.get("livemode")
        self.lookup_key = kwargs.get("lookup_key")
        self.metadata = kwargs.get("metadata")
        self.nickname = kwargs.get("nickname")
        self.product = kwargs.get("product")
        self.recurring = kwargs.get("recurring")
        self.tiers_mode = kwargs.get("tiers_mode")
        self.transform_quantity = kwargs.get("transform_quantity")
        self.type = kwargs.get("type")
        self.unit_amount = kwargs.get("unit_amount")
        self.unit_amount_decimal = kwargs.get("unit_amount_decimal")


class DatumPriceModel:
    id: str
    object: str
    billing_thresholds: None
    created: int
    metadata: Dict[str, Any]
    plan: PlanModel
    price: PriceModel
    quantity: int
    subscription: str
    tax_rates: List[Any]

    def __init__(self, **kwargs) -> None:
        self.id = kwargs.get("id")
        self.object = kwargs.get("object")
        self.billing_thresholds = kwargs.get("billing_thresholds")
        self.created = kwargs.get("created")
        self.metadata = kwargs.get("metadata")
        self.plan = kwargs.get("plan")
        self.price = kwargs.get("price")
        self.quantity = kwargs.get("quantity")
        self.subscription = kwargs.get("subscription")
        self.tax_rates = kwargs.get("tax_rates")


class ItemsPriceModel:
    object: str
    data: List[DatumPriceModel]
    has_more: bool
    total_count: int
    url: str

    def __init__(self, **kwargs) -> None:
        self.object = kwargs.get("kwargs")
        self.data = [
            DatumPriceModel(**item) for item in kwargs["data"]
        ] if "data" in kwargs else None
        self.has_more = kwargs.get("has_more")
        self.total_count = kwargs.get("total_count")
        self.url = kwargs.get("url")


class SubscriptionModel:
    id: str
    object: str
    application_fee_percent: None
    billing_cycle_anchor: int
    billing_thresholds: None
    cancel_at: None
    cancel_at_period_end: bool
    canceled_at: None
    collection_method: str
    created: int
    current_period_end: int
    current_period_start: int
    customer: str
    days_until_due: None
    default_payment_method: None
    default_source: None
    default_tax_rates: List[Any]
    discount: None
    ended_at: None
    items: ItemsModel
    latest_invoice: str
    livemode: bool
    metadata: Dict[str, Any]
    next_pending_invoice_item_invoice: None
    pause_collection: None
    pending_invoice_item_interval: None
    pending_setup_intent: None
    pending_update: None
    plan: PlanModel
    quantity: int
    schedule: None
    start_date: int
    status: str
    transfer_data: None
    trial_end: None
    trial_start: None

    def __init__(self, **kwargs) -> None:
        self.id = kwargs.get("id")
        self.object = kwargs.get("object")
        self.application_fee_percent = kwargs.get("application_fee_percent")
        self.billing_cycle_anchor = kwargs.get("billing_cycle_anchor")
        self.billing_thresholds = kwargs.get("billing_thresholds")
        self.cancel_at = kwargs.get("cancel_at")
        self.cancel_at_period_end = kwargs.get("cancel_at_period_end")
        self.canceled_at = kwargs.get("canceled_at")
        self.collection_method = kwargs.get("collection_method")
        self.created = kwargs.get("created")
        self.current_period_end = kwargs.get("current_period_end")
        self.current_period_start = kwargs.get("current_period_start")
        self.customer = kwargs.get("customer")
        self.days_until_due = kwargs.get("days_until_due")
        self.default_payment_method = kwargs.get("default_payment_method")
        self.default_source = kwargs.get("default_source")
        self.default_tax_rates = kwargs.get("default_tax_rates")
        self.discount = kwargs.get("discount")
        self.ended_at = kwargs.get("ended_at")
        self.items = kwargs.get("items")
        self.latest_invoice = kwargs.get("latest_invoice")
        self.livemode = kwargs.get("livemode")
        self.metadata = kwargs.get("metadata")
        self.next_pending_invoice_item_invoice = (
            kwargs.get("next_pending_invoice_item_invoice")
        )
        self.pause_collection = kwargs.get("pause_collection")
        self.pending_invoice_item_interval = (
            kwargs.get("pending_invoice_item_interval")
        )
        self.pending_setup_intent = kwargs.get("pending_setup_intent")
        self.pending_update = kwargs.get("pending_update")
        self.plan = kwargs.get("plan")
        self.quantity = kwargs.get("quantity")
        self.schedule = kwargs.get("schedule")
        self.start_date = kwargs.get("start_date")
        self.status = kwargs.get("status")
        self.transfer_data = kwargs.get("transfer_data")
        self.trial_end = kwargs.get("trial_end")
        self.trial_start = kwargs.get("trial_start")


class SessionModel:
    id: str
    object: str
    created: int
    customer: str
    livemode: bool
    return_url: str
    url: str

    def __init__(self, **kwargs) -> None:
        self.id = kwargs.get("id")
        self.object = kwargs.get("object")
        self.created = kwargs.get("created")
        self.customer = kwargs.get("customer")
        self.livemode = kwargs.get("livemode")
        self.return_url = kwargs.get("return_url")
        self.url = kwargs.get("url")
